# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2011-2013 Marc Deslauriers <marc.deslauriers@canonical.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from . import pytwofish
import operator
import struct


class TwofishCBC(pytwofish.Twofish):

    cbc_iv = b"\0" * 16

    def _xor(self, string, pw):
        result = b''
        for k in range(len(string)):
            result += struct.pack("B", operator.xor(ord(string[k:k + 1]),
                                  ord(pw[k:k + 1])))
        return result

    def initCBC(self, iv=b"\0" * 16):
        """Sets the CBC mode initialization vector."""
        self.cbc_iv = iv

    def encryptCBC(self, block):
        """Encrypts using CBC mode"""
        self.cbc_iv = self.encrypt(self._xor(block, self.cbc_iv))
        return self.cbc_iv

    def decryptCBC(self, block):
        decrypted_block = self._xor(self.decrypt(block), self.cbc_iv)
        self.cbc_iv = block
        return decrypted_block



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2013 Marc Deslauriers <marc.deslauriers@canonical.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import os.path
import unittest
sys.path.insert(0, os.path.realpath(os.path.join(os.path.dirname(__file__),
                                                 "..")))

from pasaffe_lib.helpers import folder_list_to_field  # noqa: E402
from pasaffe_lib.helpers import field_to_folder_list  # noqa: E402
from pasaffe_lib.helpers import folder_list_to_path  # noqa: E402
from pasaffe_lib.helpers import folder_path_to_list  # noqa: E402
from pasaffe_lib.helpers import PathEntry  # noqa: E402


class TestHelpers(unittest.TestCase):
    def test_folder_list_to_field(self):

        folder_list = [[[], ""],
                       [["foldera"], "foldera"],
                       [["folder.a"], "folder\.a"],  # noqa: W605
                       [["foldera."], "foldera\."],  # noqa: W605
                       [[".foldera"], "\.foldera"],  # noqa: W605
                       [["foldera.", "folderb."],
                         "foldera\..folderb\."],  # noqa: W605
                       [["foldera", "folderb"], "foldera.folderb"],
                       [["folder.a", "folderb"],
                         "folder\.a.folderb"],  # noqa: W605
                       [["foldera", "folder.b"],
                         "foldera.folder\.b"],  # noqa: W605
                       [["folder.a", "folder.b"],
                         "folder\.a.folder\.b"],  # noqa: W605
                       [["folder.a", "folder.b", "folder.c"],
                           "folder\.a.folder\.b.folder\.c"]]  # noqa: W605

        for (folder, field) in folder_list:
            self.assertEqual(folder_list_to_field(folder), field)

    def test_field_to_folder_list(self):

        folder_list = [["", []],
                       ["foldera", ["foldera"]],
                       ["folder\.a", ["folder.a"]],  # noqa: W605
                       ["foldera\.", ["foldera."]],  # noqa: W605
                       ["\.foldera", [".foldera"]],  # noqa: W605
                       ["foldera\..folderb\.",
                        ["foldera.", "folderb."]],  # noqa: W605
                       ["foldera.folderb", ["foldera", "folderb"]],
                       ["folder\.a.folderb",
                        ["folder.a", "folderb"]],  # noqa: W605
                       ["foldera.folder\.b",
                        ["foldera", "folder.b"]],  # noqa: W605
                       ["folder\.a.folder\.b", ["folder.a", "folder.b"]],
                       ["folder\.a.folder\.b.folder\.c",
                           ["folder.a", "folder.b", "folder.c"]]]

        for (field, folder) in folder_list:
            self.assertEqual(field_to_folder_list(field), folder)

    def test_folder_list_to_path(self):

        folder_list = [[[], "/"],
                       [["foldera"], "/foldera/"],
                       [["folder.a"], "/folder.a/"],
                       [["folder/a"], "/folder\/a/"],  # noqa: W605
                       [["foldera."], "/foldera./"],
                       [["foldera/"], "/foldera\//"],  # noqa: W605
                       [[".foldera"], "/.foldera/"],
                       [["/foldera"], "/\/foldera/"],  # noqa: W605
                       [["foldera.", "folderb."], "/foldera./folderb./"],
                       [["foldera/", "folderb/"],
                         "/foldera\//folderb\//"],  # noqa: W605
                       [["foldera", "folderb"], "/foldera/folderb/"],
                       [["folder.a", "folderb"], "/folder.a/folderb/"],
                       [["folder/a", "folderb"],
                         "/folder\/a/folderb/"],  # noqa: W605
                       [["foldera", "folder.b"], "/foldera/folder.b/"],
                       [["foldera", "folder/b"], "/foldera/folder\/b/"],
                       [["folder.a", "folder.b"], "/folder.a/folder.b/"],
                       [["folder/a", "folder/b"], "/folder\/a/folder\/b/"],
                       [["folder.a", "folder.b", "folder.c"],
                           "/folder.a/folder.b/folder.c/"],
                       [["folder/a", "folder/b", "folder/c"],
                           "/folder\/a/folder\/b/folder\/c/"]]

        for (folder, path) in folder_list:
            self.assertEqual(folder_list_to_path(folder), path)

    def test_folder_path_to_list(self):

        folder_list = [["/", []],
                       ["/foldera/", ["foldera"]],
                       ["/folder.a/", ["folder.a"]],
                       ["/folder\/a/", ["folder/a"]],  # noqa: W605
                       ["/foldera./", ["foldera."]],
                       ["/foldera\//", ["foldera/"]],  # noqa: W605
                       ["/.foldera/", [".foldera"]],
                       ["/\/foldera/", ["/foldera"]],  # noqa: W605
                       ["/foldera./folderb./", ["foldera.", "folderb."]],
                       ["/foldera\//folderb\//",
                        ["foldera/", "folderb/"]],  # noqa: W605
                       ["/foldera/folderb/", ["foldera", "folderb"]],
                       ["/folder.a/folderb/", ["folder.a", "folderb"]],
                       ["/folder\/a/folderb/", ["folder/a", "folderb"]],
                       ["/foldera/folder.b/", ["foldera", "folder.b"]],
                       ["/foldera/folder\/b/", ["foldera", "folder/b"]],
                       ["/folder.a/folder.b/", ["folder.a", "folder.b"]],
                       ["/folder\/a/folder\/b/", ["folder/a", "folder/b"]],
                       ["/folder.a/folder.b/folder.c/",
                           ["folder.a", "folder.b", "folder.c"]],
                       ["/folder\/a/folder\/b/folder\/c/",
                           ["folder/a", "folder/b", "folder/c"]]]

        for (path, folder) in folder_list:
            self.assertEqual(folder_path_to_list(path), folder)

    def test_sort_name(self):
        pathentry = PathEntry(None, None, None)

        names = [["", "zzz", -1],
                 [None, "zzz", -1],
                 ["a", "zzz", -1],
                 ["A", "zzz", -1],
                 ["A", "a", -1],
                 ["aaa", "z", -1],
                 ["aaa", "Z", -1],
                 ["aaa", "zzz", -1],
                 ["z", "zzz", -1],
                 ["aaa", "aaa", 0],
                 ["zzz", "zzz", 0],
                 [None, None, 0],
                 [None, "", 0],
                 ["", None, 0],
                 ["zzz", "z", 1],
                 ["z", "Z", 1],
                 ["zzz", "", 1],
                 ["zzz", None, 1]]

        for (first, second, result) in names:
            self.assertEqual(pathentry._sort_name(first, second), result)

    def test_sort_path(self):
        pathentry = PathEntry(None, None, None)

        paths = [[["zzz"], [], -1],
                 [["zzz"], None, -1],
                 [["a"], ["zzz"], -1],
                 [["A"], ["zzz"], -1],
                 [["A"], ["a"], -1],
                 [["aaa"], ["z"], -1],
                 [["aaa"], ["zzz"], -1],
                 [["z"], ["zzz"], -1],
                 [["a", "a"], ["a"], -1],
                 [["a", "a", "a"], ["a"], -1],
                 [["aaa"], ["aaa"], 0],
                 [["zzz"], ["zzz"], 0],
                 [None, None, 0],
                 [None, [], 0],
                 [[], None, 0],
                 [["a"], ["a", "a"], 1],
                 [["a", "a"], ["a", "a", "a"], 1],
                 [["zzz"], ["z"], 1],
                 [["zzz"], ["Z"], 1],
                 [[], ["zzz"], 1],
                 [None, ["zzz"], 1]]

        for (first, second, result) in paths:
            self.assertEqual(pathentry._sort_path(first, second), result)

    def test_sort_entries(self):
        test_entries = [["z", None],
                        ["Z", None],
                        ["a", None],
                        ["A", None],
                        ["aa", None],
                        ["a", ["a"]],
                        ["a", ["A"]],
                        ["a", ["b"]],
                        ["z", ["a", "b"]],
                        ["a", ["a", "b"]],
                        [None, ["a", "b", "c"]],
                        ["c", ["a", "b"]]]

        test_results = [["a", ["A"]],
                        [None, ["a", "b", "c"]],
                        ["a", ["a", "b"]],
                        ["c", ["a", "b"]],
                        ["z", ["a", "b"]],
                        ["a", ["a"]],
                        ["a", ["b"]],
                        ["A", None],
                        ["a", None],
                        ["aa", None],
                        ["Z", None],
                        ["z", None]]

        entries = []
        results = []

        for (name, path) in test_entries:
            new_entry = PathEntry(name, None, path)
            entries.append(new_entry)

        for (name, path) in test_results:
            new_result = PathEntry(name, None, path)
            results.append(new_result)

        # Now, sort them and check
        sorted_entries = sorted(entries)
        self.assertEqual(len(test_entries), len(results))
        self.assertEqual(len(test_entries), len(entries))
        self.assertEqual(len(test_entries), len(sorted_entries))

        for i in range(len(test_entries)):
            self.assertEqual(sorted_entries[i], results[i])


if __name__ == '__main__':
    unittest.main()



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2011 Francesco Marella <fra.marella@gmx.com>
# Copyright (C) 2012 Marc Deslauriers <marc.deslauriers@canonical.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import struct
import os
import time
from binascii import hexlify
from xml.etree import cElementTree as ET
import logging
logger = logging.getLogger('pasaffe_lib')


class KeePassX:
    def __init__(self, filename=None):
        """ Reads a KeePassX/KeePass2 XML file"""

        self.records = {}
        self.skipped = []
        self.index = 0
        self.cipher = None
        self.parent_map = None

        if filename is not None:
            self.readfile(filename)

    def _convert_time(self, timestring, tz):
        """ Converts time format"""
        if timestring:
            if tz:
                value = time.mktime(time.strptime(
                                    timestring, "%Y-%m-%dT%H:%M:%SZ"))
            else:
                value = time.mktime(time.strptime(
                                    timestring, "%Y-%m-%dT%H:%M:%S"))
        else:
            value = time.time()

        return struct.pack("<I", int(value))

    def _folder_list_to_field(self, folder_list):
        '''Converts a folder list to a folder field'''
        field = ""

        if folder_list is None:
            return field

        if folder_list == []:
            return field

        for folder in folder_list:
            if field != "":
                field += "."
            field += folder.replace(".", "\\.")
        return field

    def _get_folders(self, entry):
        '''Searches parent map to locate folders for KeePassX'''
        folder_list = []
        if entry not in self.parent_map:
            return folder_list

        parent = self.parent_map[entry]
        while True:
            if parent.tag == 'group':
                for x in list(parent):
                    if x.tag == 'title':
                        folder_list.insert(0, x.text)
                        break
            if parent not in self.parent_map:
                break
            else:
                parent = self.parent_map[parent]

        return folder_list

    def _get_folders_v2(self, entry):
        '''Searches parent map to locate folders for KeePass2'''
        folder_list = []
        if entry not in self.parent_map:
            return folder_list

        parent = self.parent_map[entry]
        while True:
            if parent.tag == 'Group':
                for x in list(parent):
                    if x.tag == 'Name':
                        folder_list.insert(0, x.text)
                        break
            if parent not in self.parent_map:
                break
            else:
                parent = self.parent_map[parent]

        # Remove the top-level 'NewDatabase' folder
        folder_list.pop(0)

        return folder_list

    def readfile(self, filename):
        """ Parses database file"""
        try:
            element = ET.parse(filename)
        except Exception:
            raise RuntimeError("Could not open %s. Aborting." % filename)

        self.parent_map = {c: p for p in element.iter() for c in p}

        if element.getroot().tag == 'database':
            for groupitem in element.findall('./group'):
                # Don't import the backups
                if groupitem.find('title').text == "Backup":
                    continue

                for pwitem in groupitem.iter('entry'):
                    uuid = os.urandom(16)
                    uuid_hex = hexlify(uuid).decode('utf-8')
                    timestamp = struct.pack("<I", int(time.time()))
                    new_entry = {1: uuid, 3: '', 4: '', 6: '',
                                 7: timestamp, 8: timestamp, 12: timestamp}

                    for x in list(pwitem):
                        if x.tag == 'title':
                            new_entry[3] = (x.text or 'Untitled item')
                        elif x.tag == 'username':
                            new_entry[4] = (x.text or '')
                        elif x.tag == 'password':
                            new_entry[6] = (x.text or '')
                        elif x.tag == 'url':
                            new_entry[13] = (x.text or '')
                        elif x.tag == 'creation':
                            new_entry[7] = self._convert_time(x.text, False)
                        elif x.tag == 'lastmod':
                            new_entry[8] = self._convert_time(x.text, False)
                            new_entry[12] = self._convert_time(x.text, False)
                        elif x.tag == 'comment':
                            new_entry[5] = (x.text or '')
                            for subelement in list(x):
                                new_entry[5] += "\n" + (subelement.tail or '')
                        else:
                            if x.tag not in self.skipped:
                                self.skipped.append(x.tag)

                    folders = self._get_folders(pwitem)
                    if folders != []:
                        new_entry[2] = self._folder_list_to_field(folders)

                    self.records[uuid_hex] = new_entry

        elif element.getroot().tag == 'KeePassFile':
            for groupitem in element.findall('./Root/Group'):
                for pwitem in groupitem.iter('Entry'):
                    uuid = os.urandom(16)
                    uuid_hex = hexlify(uuid).decode('utf-8')
                    timestamp = struct.pack("<I", int(time.time()))
                    new_entry = {1: uuid, 3: '', 4: '', 6: '',
                                 7: timestamp, 8: timestamp, 12: timestamp}

                    for x in list(pwitem):
                        if x.tag == 'Times':
                            for timesitem in list(x):
                                if timesitem.tag == 'CreationTime':
                                    new_entry[7] = self._convert_time(
                                        timesitem.text, True)
                                elif timesitem.tag == 'LastModificationTime':
                                    new_entry[8] = self._convert_time(
                                        timesitem.text, True)
                                    new_entry[12] = self._convert_time(
                                        timesitem.text, True)

                        elif x.tag == 'String':
                            for stritem in list(x):
                                if stritem.text == 'Title':
                                    new_entry[3] = (x.find('Value').text or
                                                    'Untitled item')
                                elif stritem.text == 'UserName':
                                    new_entry[4] = (x.find('Value').text or
                                                    '')
                                elif stritem.text == 'Password':
                                    new_entry[6] = (x.find('Value').text or
                                                    '')
                                elif stritem.text == 'URL':
                                    new_entry[13] = (x.find('Value').text or
                                                     '')
                                elif stritem.text == 'Notes':
                                    new_entry[5] = (x.find('Value').text or
                                                    '')
                                else:
                                    if (stritem.tag == 'Key' and
                                            stritem.text not in self.skipped):
                                        self.skipped.append(stritem.text)

                        else:
                            if x.tag not in self.skipped:
                                self.skipped.append(x.tag)

                    folders = self._get_folders_v2(pwitem)
                    if folders != []:
                        new_entry[2] = self._folder_list_to_field(folders)

                    self.records[uuid_hex] = new_entry

        else:
            raise RuntimeError("Not a valid KeePassX or KeePass2 XML file")



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2011-2013 Marc Deslauriers <marc.deslauriers@canonical.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#

'''Enhances builder connections, provides object to access glade objects'''

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import GObject, Gtk  # noqa: E402

import inspect  # noqa: E402
import functools  # noqa: E402
import logging  # noqa: E402
logger = logging.getLogger('pasaffe_lib')

from xml.etree.cElementTree import ElementTree  # noqa: E402

# this module is big so uses some conventional prefixes and postfixes
# *s list, except self.widgets is a dictionary
# *_dict dictionary
# *name string
# ele_* element in a ElementTree


# pylint: disable=R0904
# the many public methods is a feature of Gtk.Builder
class Builder(Gtk.Builder):
    ''' extra features
    connects glade defined handler to default_handler if necessary
    auto connects widget to handler with matching name or alias
    auto connects several widgets to a handler via multiple aliases
    allow handlers to lookup widget name
    logs every connection made, and any on_* not made
    '''

    def __init__(self):
        Gtk.Builder.__init__(self)
        self.widgets = {}
        self.glade_handler_dict = {}
        self.connections = []
        self._reverse_widget_dict = {}

# pylint: disable=R0201
# this is a method so that a subclass of Builder can redefine it
    def default_handler(self, handler_name, filename, *args, **kwargs):
        '''helps the apprentice guru

        glade defined handlers that do not exist come here instead.
        An apprentice guru might wonder which signal does what he wants,
        now he can define any likely candidates in glade and notice which
        ones get triggered when he plays with the project.
        this method does not appear in Gtk.Builder'''

        logger.debug('''tried to call non-existent function:%s()
        expected in %s
        args:%s
        kwargs:%s''', handler_name, filename, args, kwargs)
# pylint: enable=R0201

    def get_name(self, widget):
        ''' allows a handler to get the name (id) of a widget

        this method does not appear in Gtk.Builder'''
        return self._reverse_widget_dict.get(widget)

    def add_from_file(self, filename):
        '''parses xml file and stores wanted details'''
        Gtk.Builder.add_from_file(self, filename)

        # extract data for the extra interfaces
        tree = ElementTree()
        tree.parse(filename)

        ele_widgets = tree.getiterator("object")
        for ele_widget in ele_widgets:
            name = ele_widget.attrib['id']
            widget = self.get_object(name)

            # populate indexes - a dictionary of widgets
            self.widgets[name] = widget

            # populate a reversed dictionary
            self._reverse_widget_dict[widget] = name

            # populate connections list
            ele_signals = ele_widget.findall("signal")

            connections = [
                (name,
                 ele_signal.attrib['name'],
                 ele_signal.attrib['handler']) for ele_signal in ele_signals]

            if connections:
                self.connections.extend(connections)

        ele_signals = tree.getiterator("signal")
        for ele_signal in ele_signals:
            self.glade_handler_dict.update(
                {ele_signal.attrib["handler"]: None})

    def connect_signals(self, callback_obj):
        '''connect the handlers defined in glade

        reports successful and failed connections
        and logs call to missing handlers'''
        filename = inspect.getfile(callback_obj.__class__)
        callback_handler_dict = dict_from_callback_obj(callback_obj)
        connection_dict = {}
        connection_dict.update(self.glade_handler_dict)
        connection_dict.update(callback_handler_dict)
        for item in list(connection_dict.items()):
            if item[1] is None:
                # the handler is missing so reroute to default_handler
                handler = functools.partial(
                    self.default_handler, item[0], filename)

                connection_dict[item[0]] = handler

                # replace the run time warning
                logger.warn("expected handler '%s' in %s", item[0],
                            filename)

        # connect glade define handlers
        Gtk.Builder.connect_signals(self, connection_dict)

        # let's tell the user how we applied the glade design
        for connection in self.connections:
            widget_name, signal_name, handler_name = connection
            logger.debug("connect builder by design '%s', '%s', '%s'",
                         widget_name, signal_name, handler_name)

    def get_ui(self, callback_obj=None, by_name=True):
        '''Creates the ui object with widgets as attributes

        connects signals by 2 methods
        this method does not appear in Gtk.Builder'''

        result = UiFactory(self.widgets)

        # Hook up any signals the user defined in glade
        if callback_obj is not None:
            # connect glade define handlers
            self.connect_signals(callback_obj)

            if by_name:
                auto_connect_by_name(callback_obj, self)

        return result


# pylint: disable=R0903
# this class deliberately does not provide any public interfaces
# apart from the glade widgets
class UiFactory():
    ''' provides an object with attributes as glade widgets'''
    def __init__(self, widget_dict):
        self._widget_dict = widget_dict
        for (widget_name, widget) in list(widget_dict.items()):
            setattr(self, widget_name, widget)

        # Mangle any non-usable names (like with spaces or dashes)
        # into pythonic ones
        cannot_message = """cannot bind ui.%s, name already exists
        consider using a pythonic name instead of design name '%s'"""
        consider_message = """consider using a pythonic name instead""" + \
                           """ of design name '%s'"""

        for (widget_name, widget) in list(widget_dict.items()):
            pyname = make_pyname(widget_name)
            if pyname != widget_name:
                if hasattr(self, pyname):
                    logger.debug(cannot_message, pyname, widget_name)
                else:
                    logger.debug(consider_message, widget_name)
                    setattr(self, pyname, widget)

        def iterator():
            '''Support 'for o in self' '''
            return iter(list(widget_dict.values()))
        setattr(self, '__iter__', iterator)

    def __getitem__(self, name):
        'access as dictionary where name might be non-pythonic'
        return self._widget_dict[name]
# pylint: enable=R0903


def make_pyname(name):
    ''' mangles non-pythonic names into pythonic ones'''
    pyname = ''
    for character in name:
        if (character.isalpha() or character == '_' or
                (pyname and character.isdigit())):
            pyname += character
        else:
            pyname += '_'
    return pyname


# Until bug https://bugzilla.gnome.org/show_bug.cgi?id=652127 is fixed, we
# need to reimplement inspect.getmembers.  Gobject introspection doesn't
# play nice with it.
def getmembers(obj, check):
    members = []
    for k in dir(obj):
        try:
            attr = getattr(obj, k)
        except:  # noqa: E722
            continue
        if check(attr):
            members.append((k, attr))
    members.sort()
    return members


def dict_from_callback_obj(callback_obj):
    '''a dictionary interface to callback_obj'''
    methods = getmembers(callback_obj, inspect.ismethod)

    aliased_methods = [x[1] for x in methods if hasattr(x[1], 'aliases')]

    # a method may have several aliases
    # @alias('on_btn_foo_clicked')
    # @alias('on_tool_foo_activate')
    # on_menu_foo_activate():
    #    pass
    alias_groups = [(x.aliases, x) for x in aliased_methods]

    aliases = []
    for item in alias_groups:
        for alias in item[0]:
            aliases.append((alias, item[1]))

    dict_methods = dict(methods)
    dict_aliases = dict(aliases)

    results = {}
    results.update(dict_methods)
    results.update(dict_aliases)

    return results


def auto_connect_by_name(callback_obj, builder):
    '''finds handlers like on_<widget_name>_<signal> and connects them

    i.e. find widget,signal pair in builder and call
    widget.connect(signal, on_<widget_name>_<signal>)'''

    callback_handler_dict = dict_from_callback_obj(callback_obj)

    for item in list(builder.widgets.items()):
        (widget_name, widget) = item
        signal_ids = []
        try:
            widget_type = type(widget)
            while widget_type:
                signal_ids.extend(GObject.signal_list_ids(widget_type))
                widget_type = GObject.type_parent(widget_type)
        except RuntimeError:  # pylint wants a specific error
            pass
        signal_names = [GObject.signal_name(sid) for sid in signal_ids]

        # Now, automatically find any the user didn't specify in glade
        for sig in signal_names:
            # using convention suggested by glade
            handler_names = ["on_%s_%s" % (widget_name, sig)]

            # Using the convention that the top level window is not
            # specified in the handler name. That is use
            # on_destroy() instead of on_windowname_destroy()
            if widget is callback_obj:
                handler_names.append("on_%s" % sig)

            do_connect(item, sig, handler_names, callback_handler_dict,
                       builder.connections)

    log_unconnected_functions(callback_handler_dict, builder.connections)


def do_connect(item, signal_name, handler_names, callback_handler_dict,
               connections):
    '''connect this signal to an unused handler'''
    widget_name, widget = item

    for handler_name in handler_names:
        target = handler_name in list(callback_handler_dict.keys())
        connection = (widget_name, signal_name, handler_name)
        duplicate = connection in connections
        if target and not duplicate:
            widget.connect(signal_name, callback_handler_dict[handler_name])
            connections.append(connection)

            logger.debug("connect builder by name '%s','%s', '%s'",
                         widget_name, signal_name, handler_name)


def log_unconnected_functions(callback_handler_dict, connections):
    '''log functions like on_* that we could not connect'''

    connected_functions = [x[2] for x in connections]

    handler_names = list(callback_handler_dict.keys())
    unconnected = [x for x in handler_names if x.startswith('on_')]

    for handler_name in connected_functions:
        try:
            unconnected.remove(handler_name)
        except ValueError:
            pass

    for handler_name in unconnected:
        logger.debug("Not connected to builder '%s'", handler_name)



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2011-2013 Marc Deslauriers <marc.deslauriers@canonical.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Helpers for an Ubuntu application."""
import logging
import os
import subprocess

from . pasaffeconfig import get_data_file
from . pasaffeconfig import get_help_prefix
from . pasaffeconfig import get_help_separator

import gettext
from gettext import gettext as _
gettext.textdomain('pasaffe')


class PathEntry:
    def __init__(self, name, uuid, path):
        self.name = name
        self.uuid = uuid
        self.path = path

    def _mycmp(self, other):

        # First, we sort by path
        result = self._sort_path(self.path, other.path)
        if result != 0:
            return result

        # Then we sort by name
        result = self._sort_name(self.name, other.name)
        return result

    def __lt__(self, other):
        return self._mycmp(other) < 0

    def __gt__(self, other):
        return self._mycmp(other) > 0

    def __eq__(self, other):
        return self._mycmp(other) == 0

    def __le__(self, other):
        return self._mycmp(other) <= 0

    def __ge__(self, other):
        return self._mycmp(other) >= 0

    def __ne__(self, other):
        return self._mycmp(other) != 0

    def _lower(self, name):
        # Doing .lower() is wrong in Python 2 as it doesn't properly
        # handle certain characters in Unicode languages. Unfortunately,
        # I can't think of a better solution right now, and it will work
        # properly once we migrate to Python 3
        if name:
            return name.lower()
        else:
            return ""

    def _sort_name(self, first, second):
        # Perform a case-insensitive sort
        nocase_first = self._lower(first)
        nocase_second = self._lower(second)
        # We assume empty names are folders, so they need to
        # lose to be first in the list
        if nocase_first in (None, "") and nocase_second in (None, ""):
            return 0
        elif nocase_first < nocase_second:
            return -1
        elif nocase_first > nocase_second:
            return 1
        else:
            # If they are the same when they are case-insensitive, we now
            # want to sort in a case-sensitive way
            if first < second:
                return -1
            elif first > second:
                return 1
            else:
                return 0

    def _sort_path(self, first, second):

        # Folders should be displayed first, so they should lose to
        # entries that don't have folders
        if first in (None, []) and second in (None, []):
            return 0
        elif first in (None, []):
            return 1
        elif second in (None, []):
            return -1
        elif len(first) < len(second):
            i = 0
            for path in first:
                if not len(path):
                    return 1
                if not len(second[i]):
                    return -1
                # First test in a case insensitive way
                if self._lower(path) < self._lower(second[i]):
                    return -1
                if self._lower(path) > self._lower(second[i]):
                    return 1
                # Now try in a case-sensitive way
                if path < second[i]:
                    return -1
                if path > second[i]:
                    return 1
                i += 1
            return 1
        elif len(first) > len(second):
            i = 0
            for path in second:
                if not len(path):
                    return -1
                if not len(first[i]):
                    return 1
                # First test in a case insensitive way
                if self._lower(path) > self._lower(first[i]):
                    return -1
                if self._lower(path) < self._lower(first[i]):
                    return 1
                # Now try in a case-sensitive way
                if path > first[i]:
                    return -1
                if path < first[i]:
                    return 1
                i += 1
            return -1
        else:
            i = 0
            for path in first:
                if not len(path):
                    return 1
                if not len(second[i]):
                    return -1
                # First test in a case insensitive way
                if self._lower(path) < self._lower(second[i]):
                    return -1
                if self._lower(path) > self._lower(second[i]):
                    return 1
                # Now try in a case-sensitive way
                if path < second[i]:
                    return -1
                if path > second[i]:
                    return 1
                i += 1
            return 0

    def __repr__(self):
        return repr((self.name, self.uuid, self.path))


# Owais Lone : To get quick access to icons and stuff.
def get_media_file(media_file_name):
    media_filename = get_data_file('media', '%s' % (media_file_name,))
    if not os.path.exists(media_filename):
        media_filename = None

    return "file:///" + media_filename


class NullHandler(logging.Handler):
    def emit(self, record):
        pass


def set_up_logging(opts):
    # add a handler to prevent basicConfig
    root = logging.getLogger()
    null_handler = NullHandler()
    root.addHandler(null_handler)

    formatter = logging.Formatter("%(levelname)s:%(name)s: %(funcName)s()"
                                  " '%(message)s'")

    logger = logging.getLogger('pasaffe')
    logger_sh = logging.StreamHandler()
    logger_sh.setFormatter(formatter)
    logger.addHandler(logger_sh)

    lib_logger = logging.getLogger('pasaffe_lib')
    lib_logger_sh = logging.StreamHandler()
    lib_logger_sh.setFormatter(formatter)
    lib_logger.addHandler(lib_logger_sh)

    # Set the logging level to show debug messages.
    if opts.verbose:
        logger.setLevel(logging.DEBUG)
        logger.debug('logging enabled')
    if opts.verbose and opts.verbose > 1:
        lib_logger.setLevel(logging.DEBUG)


def get_help_uri(page=None):
    # help_uri from source tree - default language
    here = os.path.dirname(__file__)
    help_uri = os.path.abspath(os.path.join(here, '..', 'help', 'C'))
    prefix = 'ghelp:'
    separator = '#'

    if not os.path.exists(help_uri):
        # installed so use gnome help tree - user's language
        help_uri = 'pasaffe'
        prefix = get_help_prefix()
        separator = get_help_separator()

    # unspecified page is the index.page
    if page is not None:
        help_uri = '%s%s%s' % (help_uri, separator, page)

    return '%s%s' % (prefix, help_uri)


def alias(alternative_function_name):
    '''see http://www.drdobbs.com/web-development/184406073#l9'''
    def decorator(function):
        '''attach alternative_function_name(s) to function'''
        if not hasattr(function, 'aliases'):
            function.aliases = []
        function.aliases.append(alternative_function_name)
        return function
    return decorator


def folder_list_to_field(folder_list):
    '''Converts a folder list to a folder field'''
    field = ""

    if folder_list is None:
        return field

    for folder in folder_list:
        if field != "":
            field += "."
        field += folder.replace(".", "\\.")
    return field


def field_to_folder_list(field):
    '''Converts a folder field to a folder list'''

    # We need to split into folders using the "." character, but not
    # if it is escaped with a \
    folders = []

    if field == "":
        return folders

    index = 0
    location = 0
    while index < len(field):
        location = field.find(".", location + 1)

        if location == -1:
            break

        if field[location - 1] == "\\":
            continue

        folders.append(field[index:location].replace("\\", ''))
        index = location + 1

    folders.append(field[index:len(field)].replace('\\', ''))
    return folders


def folder_list_to_path(folders, index=None):
    '''Converts a folder list to a folder path'''
    if len(folders) == 0 or folders is None:
        return "/"

    if index is None:
        index = len(folders)

    folder_path = ""

    for folder in folders[0:index + 1]:
        folder_path += "/"
        folder_path += folder.replace("/", "\\/")

    folder_path += "/"

    return folder_path


def folder_path_to_list(folder_path):
    '''Converts a folder path to a folder list'''
    folders = []

    if folder_path.endswith("/"):
        folder_path = folder_path[:-1]
    if folder_path.startswith("/"):
        folder_path = folder_path[1:]

    if folder_path == '':
        return folders

    # We need to split into folders using the "/" character, but not
    # if it is escaped with a \
    index = 0
    location = 0
    while index < len(folder_path):
        location = folder_path.find("/", location + 1)

        if location == -1:
            break

        if folder_path[location - 1] == "\\":
            continue

        folders.append(folder_path[index:location].replace("\\", ''))
        index = location + 1

    folders.append(folder_path[index:len(folder_path)].replace('\\', ''))
    return folders


def confirm(prompt=None, resp=False):
    """prompts for yes or no response from the user. Returns True for yes and
    False for no.

    'resp' should be set to the default value assumed by the caller when
    user simply types ENTER.

    >>> confirm(prompt='Create Directory?', resp=True)
    Create Directory? [y]|n:
    True
    >>> confirm(prompt='Create Directory?', resp=False)
    Create Directory? [n]|y:
    False
    >>> confirm(prompt='Create Directory?', resp=False)
    Create Directory? [n]|y: y
    True

    """

    if prompt is None:
        prompt = 'Confirm'

    if resp:
        prompt = '%s [%s]|%s: ' % (prompt, 'y', 'n')
    else:
        prompt = '%s [%s]|%s: ' % (prompt, 'n', 'y')

    while True:
        ans = input(prompt)
        if not ans:
            return resp
        if ans not in ['y', 'Y', 'n', 'N']:
            print('please enter y or n.')
            continue
        if ans == 'y' or ans == 'Y':
            return True
        if ans == 'n' or ans == 'N':
            return False


def get_database_path():
    """Determines standard XDG location for database"""
    if 'XDG_DATA_HOME' in os.environ:
        basedir = os.path.join(os.environ['XDG_DATA_HOME'], 'pasaffe')
    else:
        basedir = os.path.join(os.environ['HOME'], '.local/share/pasaffe')

    if not os.path.exists(basedir):
        os.makedirs(basedir, 0o700)

    return os.path.join(basedir, 'pasaffe.psafe3')


def gen_password(number, size):
    """Generate <number> new passwords, each of size <size>"""
    command = ["apg", "-a", "1", "-n", str(number), "-M", "NCL",
               "-m", str(size),
               "-x", str(size)]
    try:
        passwords = subprocess.check_output(command).splitlines()
    except:  # noqa: E722
        print(_("error running apg"))
        passwords = None
    return passwords



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2011-2013 Marc Deslauriers <marc.deslauriers@canonical.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# THIS IS Pasaffe CONFIGURATION FILE
# YOU CAN PUT THERE SOME GLOBAL VALUE
# Do not touch unless you know what you're doing.
# you're warned :)

__all__ = ['project_path_not_found',
           'get_data_file',
           'get_data_path']

# Where your project will look for your data (for instance, images and ui
# files). By default, this is ../data, relative your trunk layout
__pasaffe_data_directory__ = '../data/'
__license__ = 'GPL-3'
__version__ = '0'
__help_prefix__ = 'help:'
__help_separator__ = '/'

import os


class project_path_not_found(Exception):
    """Raised when we can't find the project directory."""


def get_data_file(*path_segments):
    """Get the full path to a data file.

    Returns the path to a file underneath the data directory (as defined by
    `get_data_path`). Equivalent to os.path.join(get_data_path(),
    *path_segments).
    """
    return os.path.join(get_data_path(), *path_segments)


def get_data_path():
    """Retrieve pasaffe data path

    This path is by default <pasaffe_lib_path>/../data/ in trunk
    and /usr/share/pasaffe in an installed version but this path
    is specified at installation time.
    """

    # Get pathname absolute or relative.
    path = os.path.join(
        os.path.dirname(__file__), __pasaffe_data_directory__)

    abs_data_path = os.path.abspath(path)
    if not os.path.exists(abs_data_path):
        raise project_path_not_found

    return abs_data_path


def get_version():
    return __version__


def get_help_prefix():
    return __help_prefix__


def get_help_separator():
    return __help_separator__



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2011-2013 Marc Deslauriers <marc.deslauriers@canonical.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""this dialog adjusts values in gsettings
"""

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk  # noqa: E402
import logging  # noqa: E402
logger = logging.getLogger('pasaffe_lib')

from . helpers import get_help_uri  # noqa: E402
from . helpersgui import get_builder, show_uri  # noqa: E402


class PreferencesDialog(Gtk.Dialog):
    __gtype_name__ = "PreferencesDialog"

    def __new__(cls):
        """Special static method that's automatically called by Python when
        constructing a new instance of this class.

        Returns a fully instantiated PreferencesDialog object.
        """
        builder = get_builder('PreferencesPasaffeDialog')
        new_object = builder.get_object("preferences_pasaffe_dialog")
        new_object.finish_initializing(builder)
        return new_object

    def finish_initializing(self, builder):
        """Called while initializing this instance in __new__

        finish_initalizing should be called after parsing the ui definition
        and creating a PreferencesDialog object with it in order to
        finish initializing the start of the new PerferencesPasaffeDialog
        instance.

        Put your initialization code in here and leave __init__ undefined.
        """

        # Get a reference to the builder and set up the signals.
        self.builder = builder
        self.ui = builder.get_ui(self, True)

        # code for other initialization actions should be added here

    def on_btn_close_clicked(self, widget, data=None):
        self.destroy()

    def on_btn_help_clicked(self, widget, data=None):
        show_uri(self, get_help_uri('preferences'))



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2013 Marc Deslauriers <marc.deslauriers@canonical.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import os.path
import unittest
import tempfile
import shutil
import subprocess
sys.path.insert(0, os.path.realpath(os.path.join(os.path.dirname(__file__),
                                                 "..")))

from pasaffe_lib.readdb import PassSafeFile  # noqa: E402
from test_gpass_050 import TestGPass50  # noqa: E402


class TestGPassImport(TestGPass50):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.imported_db = os.path.join(self.tempdir, 'imported.psafe3')
        subprocess.call(['bin/pasaffe-import-gpass', '-q',
                         '-f', './tests/databases/gpass-050.gps',
                         '-d', self.imported_db,
                         '-y', '-p', 'pasaffe', '-m', 'pasaffe'])
        self.passfile = PassSafeFile(self.imported_db, 'pasaffe')

    def tearDown(self):
        if os.path.exists(self.tempdir):
            shutil.rmtree(self.tempdir)


if __name__ == '__main__':
    unittest.main()



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2011-2013 Marc Deslauriers <marc.deslauriers@canonical.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gio, Gtk  # noqa: E402
import logging  # noqa: E402
logger = logging.getLogger('pasaffe_lib')

from . helpers import get_help_uri  # noqa: E402
from . helpersgui import get_builder, show_uri  # noqa: E402


# This class is meant to be subclassed by PasaffeWindow.  It provides
# common functions and some boilerplate.
class Window(Gtk.Window):
    __gtype_name__ = "Window"

    # To construct a new instance of this method, the following notable
    # methods are called in this order:
    # __new__(cls)
    # __init__(self)
    # finish_initializing(self, builder)
    # __init__(self)
    #
    # For this reason, it's recommended you leave __init__ empty and put
    # your initialization code in finish_initializing

    def __new__(cls):
        """Special static method that's automatically called by Python when
        constructing a new instance of this class.

        Returns a fully instantiated BasePasaffeWindow object.
        """
        builder = get_builder('PasaffeWindow')
        new_object = builder.get_object("pasaffe_window")
        new_object.finish_initializing(builder)
        return new_object

    def finish_initializing(self, builder):
        """Called while initializing this instance in __new__

        finish_initializing should be called after parsing the UI definition
        and creating a PasaffeWindow object with it in order to finish
        initializing the start of the new PasaffeWindow instance.
        """
        # Get a reference to the builder and set up the signals.
        self.builder = builder
        self.ui = builder.get_ui(self, True)
        self.PreferencesDialog = None  # class
        self.preferences_dialog = None  # instance
        self.AboutDialog = None  # class

        self.settings = Gio.Settings.new("net.launchpad.pasaffe")
        self.settings.connect('changed', self.on_preferences_changed)

    def on_mnu_contents_activate(self, widget, data=None):
        show_uri(self, get_help_uri())

    def on_mnu_about_activate(self, widget, data=None):
        """Display the about box for pasaffe."""
        if self.AboutDialog is not None:
            about = self.AboutDialog()  # pylint: disable=E1102
            about.set_transient_for(self)
            about.run()
            about.destroy()

    def on_mnu_preferences_activate(self, widget, data=None):
        """Display the preferences window for pasaffe."""

        """ From the PyGTK Reference manual
           Say for example the preferences dialog is currently open,
           and the user chooses Preferences from the menu a second time;
           use the present() method to move the already-open dialog
           where the user can see it."""
        if self.preferences_dialog is not None:
            logger.debug('show existing preferences_dialog')
            self.preferences_dialog.present()
        elif self.PreferencesDialog is not None:
            logger.debug('create new preferences_dialog')
            self.preferences_dialog = \
                self.PreferencesDialog()  # pylint: disable=E1102
            self.preferences_dialog.set_transient_for(self)
            self.preferences_dialog.connect(
                'destroy', self.on_preferences_dialog_destroyed)
            self.preferences_dialog.show()
        # destroy command moved into dialog to allow for a help button

    def on_mnu_close_activate(self, widget, data=None):
        """Signal handler for closing the PasaffeWindow."""
        self.destroy()

    def on_destroy(self, widget, data=None):
        """Called when the PasaffeWindow is closed."""
        # Clean up code for saving application state should be added here.
        Gtk.main_quit()

    def on_preferences_changed(self, settings, key, data=None):
        logger.debug('preference changed: %s = %s' %
                     (key, str(settings.get_value(key))))
        if key == 'visible-secrets':
            self.set_show_password_status()
            treemodel, treeiter = \
                self.ui.treeview1.get_selection().get_selected()
            if treeiter is not None:
                entry_uuid = treemodel.get_value(treeiter, 2)
                if "pasaffe_treenode." not in entry_uuid:
                    self.display_data(entry_uuid,
                                      show_secrets=settings.get_boolean(key))

    def on_preferences_dialog_destroyed(self, widget, data=None):
        '''only affects gui

        logically there is no difference between the user closing,
        minimising or ignoring the preferences dialog'''
        logger.debug('on_preferences_dialog_destroyed')
        # to determine whether to create or present preferences_dialog
        self.preferences_dialog = None



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2011-2013 Marc Deslauriers <marc.deslauriers@canonical.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#

'''facade - makes pasaffe_lib package easy to refactor

while keeping its api constant'''
from . helpers import set_up_logging  # noqa: F401
from . pasaffeconfig import get_version  # noqa: F401



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2011-2013 Marc Deslauriers <marc.deslauriers@canonical.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import gettext
from gettext import gettext as _
gettext.textdomain('pasaffe')

import gi  # noqa: E402
gi.require_version('Gtk', '3.0')
from gi.repository import Gio, Gtk  # noqa: E402
from gi.repository import Gdk, Pango, GLib  # noqa: E402
import logging  # noqa: E402
import os  # noqa: E402
import webbrowser  # noqa: E402

from pasaffe_lib.Window import Window  # noqa: E402
from pasaffe.AboutPasaffeDialog import AboutPasaffeDialog  # noqa: E402
from pasaffe.EditDetailsDialog import EditDetailsDialog  # noqa: E402
from pasaffe.EditFolderDialog import EditFolderDialog  # noqa: E402
from pasaffe.PasswordEntryDialog import PasswordEntryDialog  # noqa: E402
from pasaffe.SaveChangesDialog import SaveChangesDialog  # noqa: E402
from pasaffe.NewDatabaseDialog import NewDatabaseDialog  # noqa: E402
from pasaffe.NewPasswordDialog import NewPasswordDialog  # noqa: E402
from pasaffe.PreferencesPasaffeDialog import \
    PreferencesPasaffeDialog  # noqa: E402
from pasaffe_lib.readdb import PassSafeFile  # noqa: E402
from pasaffe_lib.helpersgui import get_builder  # noqa: E402
from pasaffe_lib.helpers import folder_list_to_field  # noqa: E402
from pasaffe_lib.helpers import field_to_folder_list  # noqa: E402
from pasaffe_lib.helpers import folder_list_to_path  # noqa: E402
from pasaffe_lib.helpers import folder_path_to_list  # noqa: E402
from pasaffe_lib.helpers import PathEntry  # noqa: E402

logger = logging.getLogger('pasaffe')


# See pasaffe_lib.Window.py for more details about how this class works
class PasaffeWindow(Window):
    __gtype_name__ = "PasaffeWindow"

    def __new__(cls, database=None):
        """Special static method that's automatically called by Python when
        constructing a new instance of this class.

        Returns a fully instantiated BasePasaffeWindow object.
        """
        builder = get_builder('PasaffeWindow')
        new_object = builder.get_object("pasaffe_window")
        new_object.finish_initializing(builder, database)
        return new_object

    def finish_initializing(self, builder, database):  # pylint: disable=E1002
        """Set up the main window"""
        super(PasaffeWindow, self).finish_initializing(builder)

        self.AboutDialog = AboutPasaffeDialog
        self.EditDetailsDialog = EditDetailsDialog
        self.editdetails_dialog = None
        self.EditFolderDialog = EditFolderDialog
        self.editfolder_dialog = None
        self.PreferencesDialog = PreferencesPasaffeDialog
        self.PasswordEntryDialog = PasswordEntryDialog
        self.SaveChangesDialog = SaveChangesDialog
        self.NewDatabaseDialog = NewDatabaseDialog
        self.NewPasswordDialog = NewPasswordDialog

        self.connect("delete-event", self.on_delete_event)
        self.ui.textview1.connect("motion-notify-event",
                                  self.textview_event_handler)

        self.passfile = None
        self.is_locked = False
        self.idle_id = None
        self.clipboard_id = None
        self.last_copied = None
        self.folder_state = {}

        if database is None:
            self.database = self.settings.get_string('database-path')
        else:
            self.database = database
        self.default_database = database is None

        self.set_save_status(False)

        self.settings = Gio.Settings.new("net.launchpad.pasaffe")
        self.settings.connect('changed', self.on_preferences_changed)

        self.state = Gio.Settings.new("net.launchpad.pasaffe.state")

        # If database doesn't exists, make a new one
        if os.path.exists(self.database):
            success = self.fetch_password()
        else:
            success = self.new_database()

        if success is False:
            self.connect('event-after', Gtk.main_quit)
        else:
            self.set_window_size()
            self.set_show_password_status()
            self.display_entries()
            self.set_initial_tree_expansion()
            self.display_welcome()
            self.ui.treeview1.grab_focus()

        # Set inactivity timer
        self.set_idle_timeout()

    def on_delete_event(self, _widget, _event):
        self.save_window_size()
        return self.save_warning()

    def set_window_size(self):
        width = self.state.get_int('main-size-width')
        height = self.state.get_int('main-size-height')
        split = self.state.get_int('main-split')
        self.ui.pasaffe_window.resize(width, height)
        self.ui.hpaned1.set_position(split)

    def save_window_size(self):
        (width, height) = self.ui.pasaffe_window.get_size()
        split = self.ui.hpaned1.get_position()
        self.state.set_int('main-size-width', width)
        self.state.set_int('main-size-height', height)
        self.state.set_int('main-split', split)

    def set_entry_window_size(self):
        width = self.state.get_int('entry-size-width')
        height = self.state.get_int('entry-size-height')
        self.editdetails_dialog.ui.edit_details_dialog.resize(width, height)

    def save_entry_window_size(self):
        (width, height) = \
            self.editdetails_dialog.ui.edit_details_dialog.get_size()
        self.state.set_int('entry-size-width', width)
        self.state.set_int('entry-size-height', height)

    def set_folder_window_size(self):
        width = self.state.get_int('folder-size-width')
        height = self.state.get_int('folder-size-height')
        self.editfolder_dialog.ui.edit_folder_dialog.resize(width, height)

    def save_folder_window_size(self):
        (width, height) = \
            self.editfolder_dialog.ui.edit_folder_dialog.get_size()
        self.state.set_int('folder-size-width', width)
        self.state.set_int('folder-size-height', height)

    def save_warning(self):
        if self.get_save_status() is True:
            savechanges_dialog = self.SaveChangesDialog()
            savechanges_dialog.set_transient_for(self)
            response = savechanges_dialog.run()
            if response == Gtk.ResponseType.OK:
                self.save_db()
            elif response != Gtk.ResponseType.CLOSE:
                savechanges_dialog.destroy()
                return True
        return False

    def fetch_password(self):
        success = True
        password_dialog = self.PasswordEntryDialog()
        password_dialog.set_transient_for(self)
        while self.passfile is None:
            response = password_dialog.run()
            if response == Gtk.ResponseType.OK:
                password = password_dialog.ui.password_entry.get_text()
                try:
                    self.passfile = PassSafeFile(self.database, password)
                except ValueError:
                    password_dialog.ui.password_error_label.set_property(
                        "visible", True)
                    password_dialog.ui.password_entry.set_text("")
                    password_dialog.ui.password_entry.grab_focus()
            else:
                success = False
                break
        password_dialog.destroy()
        return success

    def new_database(self):
        success = False
        newdb_dialog = self.NewDatabaseDialog()
        newdb_dialog.set_transient_for(self)
        while success is False:
            response = newdb_dialog.run()
            if response == Gtk.ResponseType.OK:
                password_a = newdb_dialog.ui.entry1.get_text()
                password_b = newdb_dialog.ui.entry2.get_text()
                if password_a != password_b:
                    newdb_dialog.ui.error_label.set_property("visible", True)
                    newdb_dialog.ui.entry1.grab_focus()
                else:
                    self.passfile = PassSafeFile()
                    self.passfile.new_db(password_a)
                    success = True
            else:
                break

        newdb_dialog.destroy()
        return success

    def create_folders(self, folders):
        parent = None

        if not folders:
            return None

        node = self.ui.liststore1.get_iter_first()

        for folder in folders:
            if not folder:
                return parent
            found = False
            while node is not None and not found:
                if (self.ui.liststore1.get_value(node, 1) == folder) and \
                   ("pasaffe_treenode." in self.ui.liststore1.get_value(
                           node, 2)):
                    found = True
                    parent = node
                    if self.ui.liststore1.iter_has_child(node):
                        node = self.ui.liststore1.iter_children(node)
                    else:
                        break
                if not found:
                    node = self.ui.liststore1.iter_next(node)
            if not found:
                parent = self.ui.liststore1.append(parent,
                                                   ["inode-directory-symbolic",
                                                    folder,
                                                    "pasaffe_treenode." +
                                                    folder])
                node = self.ui.liststore1.iter_children(parent)
        return parent

    def _fixup_folders(self, folder_list):
        if folder_list:
            return [_("[Untitled]") if x == "" else x for x in folder_list]

        return None

    def display_entries(self):
        entries = []

        # Add empty folders first
        for folder in self.passfile.get_empty_folders():
            folder = self._fixup_folders(folder)
            entry = PathEntry("", "", folder)
            entries.append(entry)

        # Then add records
        for uuid in self.passfile.records:
            title = self.passfile.get_title(uuid)
            username = self.passfile.get_username(uuid)
            folder = self._fixup_folders(self.passfile.get_folder_list(uuid))
            # Empty names don't display properly in tree
            if title in [None, ""]:
                title = _("[Untitled]")
            elif (username not in [None, ""] and
                  self.settings.get_boolean('display-usernames') is True):
                title = title + " [" + username + "]"
            entry = PathEntry(title,
                              uuid,
                              folder)
            entries.append(entry)

        self.ui.liststore1.clear()

        # Then sort and add
        for record in sorted(entries):
            parent = self.create_folders(record.path)
            if record.name != "":
                self.ui.liststore1.append(parent,
                                          ["text-x-generic-symbolic",
                                           record.name,
                                           record.uuid])
        self.set_tree_expansion()
        self.set_menu_for_entry(False)

        # enable drag and drop
        dnd_targets = [('MY_TREE_MODEL_ROW',
                        Gtk.TargetFlags.SAME_WIDGET, 0)]

        self.ui.treeview1.enable_model_drag_source(
            Gdk.ModifierType.BUTTON1_MASK,
            dnd_targets,
            Gdk.DragAction.DEFAULT |
            Gdk.DragAction.MOVE)
        self.ui.treeview1.enable_model_drag_dest(dnd_targets,
                                                 Gdk.DragAction.DEFAULT)
        self.ui.treeview1.connect("drag_data_received",
                                  self.drag_data_received_data)

    def set_tree_expansion(self):
        folders = list(self.folder_state.copy().keys())
        folders.sort()
        for folder in folders:
            folder_iter = self.search_folder(field_to_folder_list(folder))
            if folder_iter is not None:
                path = self.ui.treeview1.get_model().get_path(folder_iter)
                if self.folder_state[folder] is True:
                    self.ui.treeview1.expand_row(path, False)
                else:
                    self.ui.treeview1.collapse_row(path)

    def set_folder_state(self, folder, state):
        folder_field = folder_list_to_field(folder)
        self.folder_state[folder_field] = state

    def set_initial_tree_expansion(self):
        entries = []

        expansion_status = self.passfile.get_tree_status()

        config = self.settings.get_string('tree-expansion')
        if config == "collapsed":
            self.ui.treeview1.collapse_all()
            return
        elif config == "expanded" or expansion_status is None:
            self.ui.treeview1.expand_all()
            return

        for folder in self.passfile.get_all_folders():
            entry = PathEntry("", "", folder)
            entries.append(entry)

        index = 0
        for record in sorted(entries):
            if index + 1 > len(expansion_status):
                return
            folder_iter = self.search_folder(record.path)
            if folder_iter is not None:
                path = self.ui.treeview1.get_model().get_path(folder_iter)
                if expansion_status[index] == "1":
                    # FIXME: For some reason, GtkTreeView will not expand
                    # a folder inside a folder that is collapsed. Need to
                    # find a workaround.
                    self.ui.treeview1.expand_row(path, False)
            index += 1

    def save_tree_expansion(self):
        entries = []
        expansion_status = ""

        for folder in self.passfile.get_all_folders():
            entry = PathEntry("", "", folder)
            entries.append(entry)

        for record in sorted(entries):
            if self.folder_state.get(folder_list_to_field(record.path),
                                     False) is True:
                expansion_status += "1"
            else:
                expansion_status += "0"

        if expansion_status == "":
            expansion_status = None

        self.passfile.set_tree_status(expansion_status)

    def drag_data_received_data(self, treeview, _context, x, y, _selection,
                                _info, _etime):
        sourcemodel, sourceiter = treeview.get_selection().get_selected()
        source_uuid = sourcemodel.get_value(sourceiter, 2)

        destmodel = treeview.get_model()
        drop_info = treeview.get_dest_row_at_pos(x, y)
        if drop_info:
            path, position = drop_info
            destiter = destmodel.get_iter(path)
            dest_uuid = destmodel.get_value(destiter, 2)

            # Ignore entries as drop destinations
            if "pasaffe_treenode." not in dest_uuid:
                return

            if ((position == Gtk.TreeViewDropPosition.INTO_OR_BEFORE) or
                    (position == Gtk.TreeViewDropPosition.INTO_OR_AFTER)):

                new_parents = self.get_folders_from_iter(destmodel, destiter)

                if "pasaffe_treenode." in source_uuid:
                    current_folders = self.get_folders_from_iter(sourcemodel,
                                                                 sourceiter)
                    parent_folders = current_folders[:-1]

                    # Bail out if the folder is dragged onto itself
                    if current_folders == new_parents:
                        return

                    # Bail out if we're trying to drag a folder into a
                    # subdirectory of itself
                    if current_folders == new_parents[:len(current_folders)]:
                        return

                    if new_parents != parent_folders:
                        new_folders = new_parents[:]
                        new_folders.append(current_folders[-1:][0])
                        old_folders = current_folders[:]

                        self.passfile.rename_folder_list(old_folders,
                                                         new_folders)

                        self.display_entries()
                        self.goto_folder(new_folders)

                        self.set_save_status(True)
                        if self.settings.get_boolean('auto-save') is True:
                            self.save_db()
                else:
                    parent_folders = self.passfile.get_folder_list(source_uuid)
                    if new_parents != parent_folders:
                        self.passfile.remove_empty_folder(new_parents)
                        self.passfile.add_empty_folder(parent_folders)
                        self.passfile.update_folder_list(source_uuid,
                                                         new_parents)

                        self.passfile.update_modification_time(source_uuid)

                        self.display_entries()
                        self.goto_uuid(source_uuid)

                        self.set_save_status(True)
                        if self.settings.get_boolean('auto-save') is True:
                            self.save_db()

                self.set_idle_timeout()
                self.update_find_results(force=True)

    def goto_uuid(self, uuid):
        item = self.search_uuid(uuid)

        if item is not None:
            treemodel = self.ui.treeview1.get_model()

            # See if we need to expand some folders
            parent = treemodel.iter_parent(item)
            while parent is not None:
                path = treemodel.get_path(parent)
                self.ui.treeview1.expand_row(path, False)
                parent = treemodel.iter_parent(parent)

            if "pasaffe_treenode." in uuid:
                self.display_folder(treemodel.get_value(item, 1))
            else:
                self.display_data(uuid)
            path = treemodel.get_path(item)
            self.ui.treeview1.scroll_to_cell(path)
            self.ui.treeview1.set_cursor(path)

    def goto_folder(self, folders):
        item = self.search_folder(folders)
        if item is not None:
            self.display_folder(self.ui.liststore1.get_value(item, 1))
            path = self.ui.treeview1.get_model().get_path(item)
            self.ui.treeview1.scroll_to_cell(path)
            self.ui.treeview1.set_cursor(path)

    def search_uuid(self, uuid, item=None, toplevel=True):
        if toplevel is True:
            item = self.ui.treeview1.get_model().get_iter_first()
        while item:
            if self.ui.liststore1.get_value(item, 2) == uuid:
                return item
            result = self.search_uuid(
                uuid, self.ui.treeview1.get_model().iter_children(item), False)
            if result:
                return result
            item = self.ui.treeview1.get_model().iter_next(item)
        return None

    def search_folder(self, folders):
        parent = None

        if not folders:
            return None

        node = self.ui.liststore1.get_iter_first()

        for folder in folders:
            if not folder:
                return parent
            found = False
            while node is not None and not found:
                if (self.ui.liststore1.get_value(node, 1) == folder) and \
                   ("pasaffe_treenode." in self.ui.liststore1.get_value(
                           node, 2)):
                    found = True
                    parent = node
                    if self.ui.liststore1.iter_has_child(node):
                        node = self.ui.liststore1.iter_children(node)
                    else:
                        break
                if not found:
                    node = self.ui.liststore1.iter_next(node)
            if not found:
                return None
        return parent

    def find_prev_iter(self, uuid, item=None, toplevel=True, prev_item=None):
        if toplevel is True:
            item = self.ui.treeview1.get_model().get_iter_first()
        if item is None:
            return prev_item, None
        while item:
            if self.ui.liststore1.get_value(item, 2) == uuid:
                return prev_item, item
            prev_item, result = self.find_prev_iter(
                uuid, self.ui.treeview1.get_model().iter_children(item),
                False, item)
            if result:
                return prev_item, result
            prev_item = item
            item = self.ui.treeview1.get_model().iter_next(item)
        return item, None

    def display_data(self, entry_uuid, show_secrets=False):
        ttt = self.get_texttagtable()
        data_buffer = Gtk.TextBuffer.new(ttt)

        # title
        data_buffer.insert(data_buffer.get_start_iter(), "\n")
        title = self.passfile.get_title(entry_uuid)
        # Fixup Empty names
        if title in [None, ""]:
            title = _("[Untitled]")

        data_buffer.insert_with_tags(data_buffer.get_end_iter(),
                                     title,
                                     ttt.lookup('title'))
        data_buffer.insert(data_buffer.get_end_iter(), "\n\n")

        # url
        if self.passfile.get_url(entry_uuid):
            data_buffer.insert_with_tags(data_buffer.get_end_iter(),
                                         _("URL:") + "\n",
                                         ttt.lookup('section'))
            data_buffer.insert_with_tags(
                data_buffer.get_end_iter(),
                self.passfile.get_url(entry_uuid),
                ttt.lookup('url'))
            data_buffer.insert(data_buffer.get_end_iter(), "\n\n")

        if show_secrets is False and \
           self.settings.get_boolean('only-passwords-are-secret') is False \
           and self.settings.get_boolean('visible-secrets') is False:
            data_buffer.insert(data_buffer.get_end_iter(),
                               _("Secrets are currently hidden."))
        else:
            # username
            data_buffer.insert_with_tags(data_buffer.get_end_iter(),
                                         _("Username:") + "\n",
                                         ttt.lookup('section'))
            data_buffer.insert(data_buffer.get_end_iter(),
                               self.passfile.get_username(entry_uuid))
            data_buffer.insert(data_buffer.get_end_iter(), "\n\n")

            # password
            data_buffer.insert_with_tags(data_buffer.get_end_iter(),
                                         _("Password:") + "\n",
                                         ttt.lookup('section'))

            if show_secrets is True or \
                    self.settings.get_boolean('visible-secrets') is True:
                data_buffer.insert_with_tags(data_buffer.get_end_iter(),
                                             self.passfile.get_password(
                                                 entry_uuid),
                                             ttt.lookup('password'))
            else:
                data_buffer.insert(data_buffer.get_end_iter(), '*****')
            data_buffer.insert(data_buffer.get_end_iter(), "\n\n")

            # notes
            if self.passfile.get_notes(entry_uuid):
                data_buffer.insert_with_tags(data_buffer.get_end_iter(),
                                             _("Notes:") + "\n",
                                             ttt.lookup('section'))
                data_buffer.insert(data_buffer.get_end_iter(),
                                   self.passfile.get_notes(entry_uuid))
                data_buffer.insert(data_buffer.get_end_iter(), "\n\n")

            # modification time
            last_updated = self.passfile.get_modification_time(entry_uuid)
            if last_updated is not None:
                data_buffer.insert_with_tags(data_buffer.get_end_iter(),
                                             _("Last updated:") + "\n",
                                             ttt.lookup('section'))
                data_buffer.insert(data_buffer.get_end_iter(),
                                   last_updated)
                data_buffer.insert(data_buffer.get_end_iter(), "\n")

            # password updated time
            pass_updated = self.passfile.get_password_time(entry_uuid)
            if pass_updated is not None:
                data_buffer.insert_with_tags(data_buffer.get_end_iter(),
                                             _("Password updated:") + "\n",
                                             ttt.lookup('section'))
                data_buffer.insert(data_buffer.get_end_iter(),
                                   pass_updated)
                data_buffer.insert(data_buffer.get_end_iter(), "\n\n")

        self.ui.textview1.set_buffer(data_buffer)

        self.set_menu_for_entry(True)

        # Now disable menus for blank entries
        if self.passfile.get_url(entry_uuid) in [None, ""]:
            self.ui.url_copy.set_sensitive(False)
            self.ui.mnu_open_url.set_sensitive(False)
            self.ui.url_copy1.set_sensitive(False)
            self.ui.open_url.set_sensitive(False)

        if self.passfile.get_username(entry_uuid) in [None, ""]:
            self.ui.username_copy.set_sensitive(False)
            self.ui.username_copy1.set_sensitive(False)
            self.ui.copy_username.set_sensitive(False)

        if self.passfile.get_password(entry_uuid) in [None, ""]:
            self.ui.password_copy.set_sensitive(False)
            self.ui.password_copy1.set_sensitive(False)
            self.ui.copy_password.set_sensitive(False)

    def display_welcome(self):
        ttt = self.get_texttagtable()
        data_buffer = Gtk.TextBuffer.new(ttt)

        data_buffer.insert(data_buffer.get_start_iter(), "\n")
        data_buffer.insert_with_tags(data_buffer.get_end_iter(),
                                     _("Welcome to Pasaffe!"),
                                     ttt.lookup('title'))
        data_buffer.insert(data_buffer.get_end_iter(), "\n\n")

        data_buffer.insert(data_buffer.get_end_iter(),
                           _("Pasaffe is an easy to use\npassword manager."))

        self.ui.textview1.set_buffer(data_buffer)

    def display_folder(self, folder_name):
        ttt = self.get_texttagtable()
        data_buffer = Gtk.TextBuffer.new(ttt)

        data_buffer.insert(data_buffer.get_start_iter(), "\n")
        data_buffer.insert_with_tags(data_buffer.get_end_iter(),
                                     folder_name,
                                     ttt.lookup('title'))
        data_buffer.insert(data_buffer.get_end_iter(), "\n\n")

        data_buffer.insert(data_buffer.get_end_iter(),
                           _("This is a folder."))

        self.ui.textview1.set_buffer(data_buffer)
        self.set_menu_for_entry(False)

    def get_texttagtable(self):
        texttagtable = Gtk.TextTagTable()
        texttag_big = Gtk.TextTag.new("title")
        texttag_big.set_property("weight", Pango.Weight.BOLD)
        texttag_big.set_property("size", 14 * Pango.SCALE)
        texttagtable.add(texttag_big)

        texttag_section = Gtk.TextTag.new("section")
        texttag_section.set_property("weight", Pango.Weight.BOLD)
        texttagtable.add(texttag_section)

        texttag_url = Gtk.TextTag.new("url")
        texttag_url.set_property("foreground", "blue")
        texttag_url.set_property("underline", Pango.Underline.SINGLE)
        texttag_url.connect("event", self.url_event_handler)
        texttagtable.add(texttag_url)

        texttag_password = Gtk.TextTag.new("password")
        texttag_password.set_property("font", 'Mono')
        texttagtable.add(texttag_password)

        return texttagtable

    def url_event_handler(self, _tag, _widget, event, _iter):
        # We also used to check event.button == 1 here, but event.button
        # doesn't seem to get set by PyGObject anymore.
        if event.type == Gdk.EventType.BUTTON_RELEASE:
            self.open_url()
        return False

    def textview_event_handler(self, textview, event):
        loc_x, loc_y = textview.window_to_buffer_coords(
            Gtk.TextWindowType.WIDGET, int(event.x), int(event.y))

        # Fix dumb Gtk 3.20 API change
        try:
            itera = textview.get_iter_at_location(loc_x, loc_y)
            tags = itera.get_tags()
        except AttributeError:
            (_over, itera) = textview.get_iter_at_location(loc_x, loc_y)
            tags = itera.get_tags()

        cursor = Gdk.Cursor.new(Gdk.CursorType.XTERM)
        for tag in tags:
            if tag.get_property('name') == 'url':
                cursor = Gdk.Cursor.new(Gdk.CursorType.HAND2)
                break
        textview.get_window(Gtk.TextWindowType.TEXT).set_cursor(cursor)
        return False

    def open_url(self):
        url = None
        treemodel, treeiter = self.ui.treeview1.get_selection().get_selected()
        if treeiter is not None:
            entry_uuid = treemodel.get_value(treeiter, 2)

            # Bail out of we're a folder
            if "pasaffe_treenode." in entry_uuid:
                return

            url = self.passfile.get_url(entry_uuid)
        if url is not None:
            if not url.startswith('http://') and \
               not url.startswith('https://'):
                url = 'http://' + url
            webbrowser.open(url)

    def on_treeview1_cursor_changed(self, treeview):
        self.set_idle_timeout()
        selection = treeview.get_selection()
        if selection is not None:
            treemodel, treeiter = selection.get_selected()
            if treemodel is not None and treeiter is not None:
                entry_uuid = treemodel.get_value(treeiter, 2)
                if "pasaffe_treenode." in entry_uuid:
                    self.display_folder(treemodel.get_value(treeiter, 1))
                else:
                    self.display_data(entry_uuid)

                # Reset the show password button and menu item
                self.ui.display_secrets.set_active(False)
                self.ui.mnu_display_secrets.set_active(False)

    def on_treeview1_button_press_event(self, treeview, event):
        if event.button == 3:
            loc_x = int(event.x)
            loc_y = int(event.y)
            event_time = event.time
            pthinfo = treeview.get_path_at_pos(loc_x, loc_y)
            if pthinfo is not None:
                path, col, _cellx, _celly = pthinfo
                treeview.grab_focus()
                treeview.set_cursor(path, col, 0)

                if not Gtk.check_version(3, 22, 0):
                    self.ui.menu_popup.popup_at_pointer(event)
                else:
                    self.ui.menu_popup.popup(None, None, None, None, 3,
                                             event_time)

    def add_entry(self):
        self.disable_idle_timeout()

        # Make sure dialog isn't already open
        if self.editdetails_dialog is not None:
            self.editdetails_dialog.present()
            return

        uuid_hex = self.passfile.new_entry()

        treemodel, treeiter = self.ui.treeview1.get_selection().get_selected()

        folder = None
        if treeiter is not None:
            folder = self.get_folders_from_iter(treemodel, treeiter)

        if folder is not None:
            # expand folder
            path = self.ui.treeview1.get_model().get_path(treeiter)
            self.ui.treeview1.expand_row(path, False)

            self.passfile.update_folder_list(uuid_hex, folder)

        response = self.edit_entry(uuid_hex, True)
        if response != Gtk.ResponseType.OK:
            self.delete_entry(uuid_hex, save=False)
        else:
            self.display_entries()
            self.goto_uuid(uuid_hex)
            self.set_save_status(True)
            if self.settings.get_boolean('auto-save') is True:
                self.save_db()
        self.set_idle_timeout()
        self.update_find_results(force=True)

    def add_folder(self):
        self.disable_idle_timeout()

        # Make sure dialog isn't already open
        if self.editfolder_dialog is not None:
            self.editfolder_dialog.present()
            return

        # Get currently selected folder
        treemodel, treeiter = self.ui.treeview1.get_selection().get_selected()
        if treeiter is not None:
            entry_uuid = treemodel.get_value(treeiter, 2)
            if "pasaffe_treenode." not in entry_uuid:
                treeiter = None

        # TODO: make sure folder name is unique in same level

        new_iter = self.ui.liststore1.append(treeiter,
                                             ['inode-directory-symbolic',
                                              _('New Folder'),
                                              "pasaffe_treenode.New Folder"])
        if treeiter is not None:
            path = self.ui.treeview1.get_model().get_path(treeiter)
            self.ui.treeview1.expand_row(path, False)

        self.ui.treeview1.get_selection().select_iter(new_iter)
        self.display_folder(self.ui.liststore1.get_value(new_iter, 1))

        new_folder = self.get_folders_from_iter(treemodel, new_iter)
        response = self.edit_folder(self.ui.liststore1, new_iter, True)

        if response != Gtk.ResponseType.OK:
            self.delete_folder(new_folder, save=False)

    def clone_entry(self, entry_uuid):
        record_list = (2, 3, 4, 5, 6, 13)
        self.disable_idle_timeout()

        # Make sure dialog isn't already open
        if self.editdetails_dialog is not None:
            self.editdetails_dialog.present()
            return

        uuid_hex = self.passfile.new_entry()

        for record_type in record_list:
            if record_type in self.passfile.records[entry_uuid]:
                self.passfile.records[uuid_hex][record_type] = \
                    self.passfile.records[entry_uuid][record_type]

        response = self.edit_entry(uuid_hex)
        if response != Gtk.ResponseType.OK:
            self.delete_entry(uuid_hex, save=False)
        else:
            self.display_entries()
            self.goto_uuid(uuid_hex)
            self.set_save_status(True)
            if self.settings.get_boolean('auto-save') is True:
                self.save_db()
        self.set_idle_timeout()
        self.update_find_results(force=True)

    def remove_entry(self):
        treemodel, treeiter = self.ui.treeview1.get_selection().get_selected()
        if treeiter is not None:
            entry_uuid = treemodel.get_value(treeiter, 2)
            entry_name = treemodel.get_value(treeiter, 1)

            information = \
                _('<big><b>Are you sure you wish to'
                    ' remove "%s"?</b></big>\n\n') % \
                GLib.markup_escape_text(entry_name)
            information += _('Contents of the entry will be lost.\n')

            info_dialog = Gtk.MessageDialog(
                transient_for=self,
                modal=True,
                message_type=Gtk.MessageType.QUESTION,
                buttons=Gtk.ButtonsType.YES_NO)
            info_dialog.set_markup(information)
            result = info_dialog.run()
            info_dialog.destroy()

            if result == Gtk.ResponseType.YES:
                self.delete_entry(entry_uuid)

    def remove_folder(self):
        treemodel, treeiter = self.ui.treeview1.get_selection().get_selected()
        if treeiter is not None:
            folder_name = treemodel.get_value(treeiter, 1)

            information = \
                _('<big><b>Are you sure you wish'
                  ' to remove folder "%s"?</b></big>\n\n') % \
                GLib.markup_escape_text(folder_name)
            information += _('All entries in this folder will be lost.\n')

            info_dialog = Gtk.MessageDialog(
                transient_for=self,
                modal=True,
                message_type=Gtk.MessageType.QUESTION,
                buttons=Gtk.ButtonsType.YES_NO)
            info_dialog.set_markup(information)
            result = info_dialog.run()
            info_dialog.destroy()

            if result == Gtk.ResponseType.YES:
                folder = self.get_folders_from_iter(treemodel, treeiter)
                self.delete_folder(folder)

    def populate_folders(self, liststore, combobox, default=None):

        folders = [["/"]]
        for folder in self.passfile.get_all_folders():
            for index in range(len(folder)):
                folder_path = [folder_list_to_path(folder, index)]
                if folder_path not in folders:
                    folders.append(folder_path)

        folders.sort()

        for folder in folders:
            liststore.append(folder)

        if default is not None:
            item = self.search_folder_ui(liststore,
                                         folder_list_to_path(default))
            if item is not None:
                combobox.set_active_iter(item)

    def search_folder_ui(self, liststore, folder):
        item = liststore.get_iter_first()
        while item:
            if liststore.get_value(item, 0) == folder:
                return item
            item = liststore.iter_next(item)
        return None

    def edit_entry(self, entry_uuid, new_entry=False):
        if "pasaffe_treenode." in entry_uuid:
            return None
        record_dict = {2: 'folder_entry',
                       3: 'name_entry',
                       4: 'username_entry',
                       5: 'notes_buffer',
                       6: 'password_entry',
                       13: 'url_entry'}

        # Make sure dialog isn't already open
        if self.editdetails_dialog is not None:
            self.editdetails_dialog.present()
            return

        if self.EditDetailsDialog is not None:
            self.disable_idle_timeout()
            self.editdetails_dialog = self.EditDetailsDialog()
            if new_entry is True:
                title = self.editdetails_dialog.builder.get_object('title')
                title.set_markup(_("<big><b>New entry</b></big>"))

            for record_type, widget_name in list(record_dict.items()):
                # Handle folders separately
                if record_type == 2:
                    liststore = \
                        self.editdetails_dialog.builder.get_object(
                            'liststore1')
                    combobox = \
                        self.editdetails_dialog.builder.get_object(
                            'folder_combo')
                    if self.passfile.get_folder_list(entry_uuid):
                        self.populate_folders(
                            liststore,
                            combobox,
                            self.passfile.get_folder_list(entry_uuid))
                    else:
                        self.populate_folders(liststore, combobox, [])
                elif record_type == 3:
                    self.editdetails_dialog.builder.get_object(
                        widget_name).set_text(
                            self.passfile.get_title(entry_uuid))
                elif record_type in self.passfile.records[entry_uuid]:
                    self.editdetails_dialog.builder.get_object(
                        widget_name).set_text(
                            self.passfile.records[entry_uuid][record_type])

            # Set a Mono font so certain characters don't look alike
            # How do I do this in the glade file?
            password_entry = self.editdetails_dialog.builder.get_object(
                'password_entry')
            password_entry.modify_font(Pango.FontDescription('Mono'))

            self.set_entry_window_size()
            self.editdetails_dialog.set_transient_for(self)
            response = self.editdetails_dialog.run()

            data_changed = False
            tree_changed = False

            if response == Gtk.ResponseType.OK:
                for record_type, widget_name in list(record_dict.items()):
                    # Get the new value
                    if record_type == 2:
                        combo = self.editdetails_dialog.builder.get_object(
                            'folder_combo')
                        combo_iter = combo.get_active_iter()
                        if combo_iter is not None:
                            new_value = folder_path_to_list(
                                combo.get_model()[combo_iter][0])
                        else:
                            new_value = folder_path_to_list(
                                combo.get_child().get_text())
                    elif record_type == 5:
                        new_value = self.editdetails_dialog.builder.get_object(
                            widget_name).get_text(
                                self.editdetails_dialog.builder.get_object(
                                    widget_name).get_start_iter(),
                                self.editdetails_dialog.builder.get_object(
                                    widget_name).get_end_iter(), True)
                    else:
                        new_value = self.editdetails_dialog.builder.get_object(
                            widget_name).get_text()

                    # Now do something with it
                    if record_type == 2:
                        old_folder = self.passfile.get_folder_list(entry_uuid)
                        self.passfile.remove_empty_folder(new_value)
                        if old_folder != new_value:
                            self.passfile.add_empty_folder(old_folder)
                            self.passfile.update_folder_list(
                                entry_uuid, new_value)
                            data_changed = True
                            tree_changed = True
                    elif ((record_type in [5, 13]) and
                          new_value == "" and
                          record_type in self.passfile.records[entry_uuid]):
                        del self.passfile.records[entry_uuid][record_type]
                        data_changed = True
                    elif self.passfile.records[entry_uuid].get(
                            record_type, "") != new_value:
                        data_changed = True
                        self.passfile.records[entry_uuid][record_type] = \
                            new_value
                        # Reset the entire tree on name and path changes
                        if record_type in [2, 3, 4]:
                            tree_changed = True

                        # Update the password changed date
                        if record_type == 6:
                            self.passfile.update_password_time(entry_uuid)

                if data_changed is True:
                    self.set_save_status(True)
                    self.passfile.update_modification_time(entry_uuid)
                    if self.settings.get_boolean('auto-save') is True:
                        self.save_db()

            self.save_entry_window_size()
            self.editdetails_dialog.destroy()
            self.editdetails_dialog = None

            if tree_changed is True:
                self.display_entries()
                self.goto_uuid(entry_uuid)
            else:
                # Update the right pane only if it's still the one
                # currently selected
                treemodel, treeiter = \
                    self.ui.treeview1.get_selection().get_selected()
                if treeiter is not None and treemodel.get_value(
                        treeiter, 2) == entry_uuid:
                    self.display_data(entry_uuid)

            self.set_idle_timeout()
            self.update_find_results(force=True)
            return response

    def edit_folder(self, treemodel, treeiter, new_folder=False):

        # Make sure dialog isn't already open
        if self.editfolder_dialog is not None:
            self.editfolder_dialog.present()
            return

        if self.EditFolderDialog is not None:
            self.disable_idle_timeout()
            self.editfolder_dialog = self.EditFolderDialog()

            if new_folder is True:
                title = self.editfolder_dialog.builder.get_object('title')
                title.set_markup(_("<big><b>New folder</b></big>"))

            folder_name = treemodel.get_value(treeiter, 1)
            if folder_name == _("[Untitled]"):
                folder_name = ""
            self.editfolder_dialog.ui.folder_name_entry.set_text(folder_name)
            self.editfolder_dialog.ui.folder_name_entry.select_region(0, -1)

            liststore = self.editfolder_dialog.builder.get_object('liststore1')
            combobox = self.editfolder_dialog.builder.get_object(
                'folder_combo')
            parent_folder = self.get_folders_from_iter(
                treemodel, treeiter)[:-1]
            self.populate_folders(liststore, combobox, parent_folder)

            self.set_folder_window_size()
            self.editfolder_dialog.set_transient_for(self)
            response = self.editfolder_dialog.run()
            if response == Gtk.ResponseType.OK:
                new_name = \
                    self.editfolder_dialog.ui.folder_name_entry.get_text()

                combo_iter = combobox.get_active_iter()
                if combo_iter is not None:
                    new_parent = folder_path_to_list(
                        combobox.get_model()[combo_iter][0])
                else:
                    new_parent = []

                if new_name != folder_name or new_parent != parent_folder:

                    # TODO: make sure new_name is unique in the same level
                    new_folders = new_parent[:]
                    new_folders.append(new_name)

                    old_folders = parent_folder[:]
                    old_folders.append(folder_name)

                    if new_folder is False:
                        self.passfile.rename_folder_list(
                            old_folders, new_folders)
                    else:
                        self.passfile.add_empty_folder(new_folders)

                    self.display_entries()
                    self.goto_folder(new_folders)

                    self.set_save_status(True)
                    if self.settings.get_boolean('auto-save') is True:
                        self.save_db()

            self.save_folder_window_size()
            self.editfolder_dialog.destroy()
            self.editfolder_dialog = None

            self.set_idle_timeout()
            self.update_find_results(force=True)
            return response

    def get_folders_from_iter(self, treemodel, treeiter):
        folders = []

        if treemodel is None or treeiter is None:
            return folders

        uuid = treemodel.get_value(treeiter, 2)
        if "pasaffe_treenode." in uuid:
            folders.append(treemodel.get_value(treeiter, 1))

        parent = treemodel.iter_parent(treeiter)
        while parent is not None:
            folders.insert(0, treemodel.get_value(parent, 1))
            parent = treemodel.iter_parent(parent)

        return folders

    def delete_entry(self, entry_uuid, save=True):
        self.set_idle_timeout()

        item = self.search_uuid(entry_uuid)
        if item:
            new_item = self.ui.treeview1.get_model().iter_next(item)

            if new_item is None:
                # No more items in the current level, try and get the parent
                new_item, _item = self.find_prev_iter(entry_uuid)

            self.ui.liststore1.remove(item)

        # Delete the entry, and add the folder to the empty list
        folder = self.passfile.get_folder_list(entry_uuid)
        self.passfile.delete_entry(entry_uuid)
        self.passfile.add_empty_folder(folder)

        if item:
            if new_item is None:
                new_item = self.ui.treeview1.get_model().get_iter_first()

            if new_item != 0 and new_item is not None:
                self.ui.treeview1.get_selection().select_iter(new_item)

        if save is True:
            self.set_save_status(True)
            if self.settings.get_boolean('auto-save') is True:
                self.save_db()

        treemodel, treeiter = self.ui.treeview1.get_selection().get_selected()
        if treeiter is not None:
            entry_uuid = treemodel.get_value(treeiter, 2)
            if "pasaffe_treenode." in entry_uuid:
                self.display_folder(treemodel.get_value(treeiter, 1))
            else:
                self.display_data(entry_uuid)
        else:
            self.display_welcome()

        self.update_find_results(force=True)

    def delete_folder(self, folders, save=True):
        self.set_idle_timeout()

        item = self.search_folder(folders)
        if item:
            self.passfile.delete_folder(folders)

            self.display_entries()
            parent_folder = folders[:-1]
            # TODO: if top level, switch to next iter
            if parent_folder != []:
                self.goto_folder(parent_folder)

        if save is True:
            self.set_save_status(True)
            if self.settings.get_boolean('auto-save') is True:
                self.save_db()

        treemodel, treeiter = self.ui.treeview1.get_selection().get_selected()
        if treeiter is not None:
            entry_uuid = treemodel.get_value(treeiter, 2)
            if "pasaffe_treenode." in entry_uuid:
                self.display_folder(treemodel.get_value(treeiter, 1))
            else:
                self.display_data(entry_uuid)
        else:
            self.display_welcome()

        self.update_find_results(force=True)

    def model_get_iter_last(self, model, parent=None):
        """Returns a Gtk.TreeIter to the last row or None if there aren't any rows.
        If parent is None, returns a Gtk.TreeIter to the last root row."""
        nchild = model.iter_n_children(parent)
        return nchild and model.iter_nth_child(parent, nchild - 1)

    def on_treeview1_row_activated(self, treeview, _path, _view_column):
        treemodel, treeiter = treeview.get_selection().get_selected()
        entry_uuid = treemodel.get_value(treeiter, 2)
        if "pasaffe_treenode." in entry_uuid:
            # Toggle expanded state of folder
            folder = self.get_folders_from_iter(treemodel, treeiter)
            folder_field = folder_list_to_field(folder)
            if folder_field in self.folder_state and \
               self.folder_state[folder_field] is True:
                self.collapse_folder(folder)
            else:
                self.expand_folder(folder)
        else:
            if self.settings.get_string('double-click') == "copies":
                # Copy password
                self.copy_selected_entry_item(6)
            else:
                self.edit_entry(entry_uuid)

    def collapse_folder(self, folder):
        self.set_folder_state(folder, False)
        self.set_tree_expansion()

    def expand_folder(self, folder):
        self.set_folder_state(folder, True)
        self.set_tree_expansion()

    def on_treeview1_row_expanded(self, treeview, treeiter, _path):
        treemodel = treeview.get_model()
        folder = self.get_folders_from_iter(treemodel, treeiter)
        self.set_folder_state(folder, True)

    def on_treeview1_row_collapsed(self, treeview, treeiter, _path):
        treemodel = treeview.get_model()
        folder = self.get_folders_from_iter(treemodel, treeiter)
        self.set_folder_state(folder, False)

    def on_treeview1_key_pressed(self, treeview, event):
        treemodel, treeiter = self.ui.treeview1.get_selection().get_selected()
        if treeiter is None:
            return
        entry_uuid = treemodel.get_value(treeiter, 2)
        if event.keyval in [Gdk.KEY_Return, Gdk.KEY_KP_Enter]:
            if "pasaffe_treenode." in entry_uuid:
                return
            else:
                self.edit_entry(entry_uuid)
                return True
        if event.keyval == Gdk.KEY_Left:
            if "pasaffe_treenode." in entry_uuid:
                folder = self.get_folders_from_iter(treemodel, treeiter)
                folder_field = folder_list_to_field(folder)
                if folder_field in self.folder_state and \
                   self.folder_state[folder_field] is True:
                    self.collapse_folder(folder)
                    return True
                folders = self.get_folders_from_iter(treemodel, treeiter)[:-1]
            else:
                folders = self.get_folders_from_iter(treemodel, treeiter)
            if folders:
                # Select parent folder
                self.goto_folder(folders)
            return True
        if event.keyval == Gdk.KEY_Right:
            if "pasaffe_treenode." in entry_uuid:
                folder = self.get_folders_from_iter(treemodel, treeiter)
                folder_field = folder_list_to_field(folder)
                if folder_field not in self.folder_state or \
                   self.folder_state[folder_field] is False:
                    self.expand_folder(folder)
                    return True
                # Select first child item
                iterchild = \
                    self.ui.treeview1.get_model().iter_children(treeiter)
                if iterchild:
                    uuid = treemodel.get_value(iterchild, 2)
                    self.goto_uuid(uuid)
                return True

    def save_db(self):
        if self.get_save_status() is True:
            self.save_tree_expansion()
            self.passfile.writefile(self.database, backup=True)
            self.set_save_status(False)

    def on_save_clicked(self, _toolbutton):
        self.set_idle_timeout()
        self.save_db()

    def on_mnu_save_activate(self, _menuitem):
        self.set_idle_timeout()
        self.save_db()

    def on_mnu_close_activate(self, _menuitem):
        self.disable_idle_timeout()
        if self.settings.get_boolean('auto-save') is True:
            self.save_db()
        if self.save_warning() is False:
            Gtk.main_quit()
        else:
            self.set_idle_timeout()

    def on_mnu_clone_activate(self, _menuitem):
        treemodel, treeiter = self.ui.treeview1.get_selection().get_selected()
        if treeiter is not None:
            entry_uuid = treemodel.get_value(treeiter, 2)
            # TODO: what happens when we clone a folder?
            if "pasaffe_treenode." in entry_uuid:
                return
            else:
                self.clone_entry(entry_uuid)

    def on_username_copy_activate(self, _menuitem):
        self.copy_selected_entry_item(4)

    def on_password_copy_activate(self, _menuitem):
        self.copy_selected_entry_item(6)

    def on_url_copy_activate(self, _menuitem):
        self.copy_selected_entry_item(13)

    def on_copy_username_clicked(self, _toolbutton):
        self.copy_selected_entry_item(4)

    def on_copy_password_clicked(self, _toolbutton):
        self.copy_selected_entry_item(6)

    def on_display_secrets_toggled(self, toolbutton):
        is_active = toolbutton.get_active()
        self.display_secrets(is_active)
        self.ui.mnu_display_secrets.set_active(is_active)

    def on_mnu_display_secrets_toggled(self, menuitem):
        is_active = menuitem.get_active()
        self.display_secrets(is_active)
        self.ui.display_secrets.set_active(is_active)

    def display_secrets(self, display=True):
        self.set_idle_timeout()
        treemodel, treeiter = self.ui.treeview1.get_selection().get_selected()
        if treeiter is not None:
            entry_uuid = treemodel.get_value(treeiter, 2)

            # Bail out of we're a folder
            if "pasaffe_treenode." in entry_uuid:
                return

            self.display_data(entry_uuid, display)

    def copy_selected_entry_item(self, item):
        self.set_idle_timeout()
        treemodel, treeiter = self.ui.treeview1.get_selection().get_selected()
        if treeiter is not None:
            entry_uuid = treemodel.get_value(treeiter, 2)

            # Bail out of we're a folder
            if "pasaffe_treenode." in entry_uuid:
                return

            if item in self.passfile.records[entry_uuid]:
                for atom in [Gdk.SELECTION_CLIPBOARD, Gdk.SELECTION_PRIMARY]:
                    clipboard = Gtk.Clipboard.get(atom)
                    value = self.passfile.records[entry_uuid][item]
                    length = len(value.encode('utf-8'))
                    clipboard.set_text(value, length)
                    self.last_copied = value
                    clipboard.store()
                self.set_clipboard_timeout()

    def on_mnu_add_entry_activate(self, _menuitem):
        self.add_entry()

    def on_mnu_add_folder_activate(self, _menuitem):
        self.add_folder()

    def on_mnu_edit1_activate(self, _menuitem):
        treemodel, treeiter = self.ui.treeview1.get_selection().get_selected()
        if treeiter is not None:
            entry_uuid = treemodel.get_value(treeiter, 2)
            if "pasaffe_treenode." in entry_uuid:
                self.edit_folder(treemodel, treeiter)
            else:
                self.edit_entry(entry_uuid)

    def on_mnu_delete_activate(self, _menuitem):
        treemodel, treeiter = self.ui.treeview1.get_selection().get_selected()
        if treeiter is not None:
            entry_uuid = treemodel.get_value(treeiter, 2)
            if "pasaffe_treenode." in entry_uuid:
                self.remove_folder()
            else:
                self.remove_entry()

    def on_mnu_lock_activate(self, _menuitem):
        self.lock_screen()

    def on_mnu_info_activate(self, _menuitem):
        header = _('<big><b>Database Information</b></big>\n\n')
        information = _('Number of entries: %s\n') % \
            len(self.passfile.records)
        information += '\n'
        if self.passfile.get_saved_name():
            information += _('Last saved by: %s\n') % \
                self.passfile.get_saved_name()
        if self.passfile.get_saved_host():
            information += _('Last saved on host: %s\n') % \
                self.passfile.get_saved_host()
            information += _('Last save date: %s\n') % \
                self.passfile.get_saved_date_string()
        information += '\n'
        information += _('Database version: %s\n') % \
            self.passfile.get_database_version_string()
        if self.passfile.get_saved_application():
            information += _('Application used: %s\n') % \
                self.passfile.get_saved_application()
        information += '\n'
        information += _('Database location:\n%s\n') % self.database
        information = header + GLib.markup_escape_text(information)

        info_dialog = Gtk.MessageDialog(transient_for=self,
                                        modal=True,
                                        message_type=Gtk.MessageType.INFO,
                                        buttons=Gtk.ButtonsType.OK)
        info_dialog.set_markup(information)
        info_dialog.run()
        info_dialog.destroy()

    def on_mnu_open_url_activate(self, _menuitem):
        self.open_url()

    def on_mnu_find_toggled(self, menuitem):
        is_active = menuitem.get_active()
        self.show_find(is_active)
        self.ui.toolbar_find.set_active(is_active)

    def on_toolbar_find_toggled(self, toolbutton):
        is_active = toolbutton.get_active()
        self.show_find(is_active)
        self.ui.mnu_find.set_active(is_active)

    def on_find_btn_close_clicked(self, _button):
        self.show_find(False)
        self.ui.toolbar_find.set_active(False)
        self.ui.mnu_find.set_active(False)

    def on_find_btn_prev_clicked(self, _button):
        self.update_find_results()
        self.goto_next_find_result(backwards=True)

    def on_find_btn_next_clicked(self, _button):
        self.update_find_results()
        self.goto_next_find_result()

    def update_find_results(self, force=False):

        if self.ui.find_box.get_property("visible") is False:
            return

        find = self.ui.find_entry.get_text()

        # Don't crash with stupid "bogus escape (end of line)" error
        while find[-1:] == '\\':
            find = find[:-1]

        self.passfile.update_find_results(find, force)

    def goto_next_find_result(self, backwards=False):

        uuid_hex = self.passfile.get_next_find_result(backwards)

        if uuid_hex is not None:
            self.goto_uuid(uuid_hex)

    def on_find_entry_activate(self, _entry):
        self.update_find_results()
        self.goto_next_find_result()

    def show_find(self, show):

        if self.ui.find_box.get_property("visible") == show:
            return

        if show is True:
            self.ui.find_entry.set_text("")
            self.find_value = ""
            self.ui.find_box.set_property("visible", True)
            self.ui.find_entry.grab_focus()
        else:
            self.ui.find_box.set_property("visible", False)
            self.find_results = []
            self.find_results_index = None

    def on_open_url_clicked(self, _toolbutton):
        self.open_url()

    def on_mnu_chg_password_activate(self, _menuitem):
        success = False
        newpass_dialog = self.NewPasswordDialog()
        newpass_dialog.set_transient_for(self)

        while success is False:
            response = newpass_dialog.run()
            if response == Gtk.ResponseType.OK:
                old_password = newpass_dialog.ui.pass_entry1.get_text()
                password_a = newpass_dialog.ui.pass_entry2.get_text()
                password_b = newpass_dialog.ui.pass_entry3.get_text()
                if password_a != password_b:
                    newpass_dialog.ui.label3.set_text(
                        _("Passwords don't match! Please try again."))
                    newpass_dialog.ui.label3.set_property("visible", True)
                    newpass_dialog.ui.pass_entry2.grab_focus()
                elif password_a == '':
                    newpass_dialog.ui.label3.set_text(
                        _("New password cannot be blank! Please try again."))
                    newpass_dialog.ui.label3.set_property("visible", True)
                    newpass_dialog.ui.pass_entry2.grab_focus()
                elif not self.passfile.check_password(old_password):
                    newpass_dialog.ui.label3.set_text(
                        _("Old password is invalid! Please try again."))
                    newpass_dialog.ui.label3.set_property("visible", True)
                    newpass_dialog.ui.pass_entry1.grab_focus()
                else:
                    self.passfile.new_keys(password_a)
                    self.set_save_status(True)
                    self.save_db()
                    success = True
            else:
                break

        newpass_dialog.destroy()

    def lock_screen(self):
        self.disable_idle_timeout()
        self.clear_clipboard()
        self.is_locked = True
        self.ui.pasaffe_vbox.reparent(self.ui.empty_window)
        self.ui.lock_vbox.reparent(self.ui.pasaffe_window)
        self.set_menu_sensitive(False)
        self.ui.lock_unlock_button.grab_focus()

    def on_pasaffe_window_delete_event(self, _window, event):
        # Pasaffe window is closing
        self.clear_clipboard()

    def on_lock_unlock_button_clicked(self, _button):
        success = False
        password_dialog = self.PasswordEntryDialog()
        password_dialog.set_transient_for(self)
        password_dialog.set_modal(True)
        while success is False:
            response = password_dialog.run()
            if response == Gtk.ResponseType.OK:
                password = password_dialog.ui.password_entry.get_text()
                success = self.passfile.check_password(password)
                if success is False:
                    password_dialog.ui.password_error_label.set_property(
                        "visible", True)
                    password_dialog.ui.password_entry.set_text("")
                    password_dialog.ui.password_entry.grab_focus()
            else:
                password_dialog.destroy()
                return
        password_dialog.destroy()
        self.ui.lock_vbox.reparent(self.ui.lock_window)
        self.ui.pasaffe_vbox.reparent(self.ui.pasaffe_window)
        self.set_menu_sensitive(True)
        self.is_locked = False
        self.set_idle_timeout()

    def on_lock_quit_button_clicked(self, _button):
        if self.save_warning() is False:
            Gtk.main_quit()
            return

    def set_menu_sensitive(self, status):
        # There's got to be a better way to do this
        self.ui.mnu_lock.set_sensitive(status)
        self.ui.mnu_chg_password.set_sensitive(status)
        self.ui.mnu_add_entry.set_sensitive(status)
        self.ui.mnu_add_folder.set_sensitive(status)
        self.ui.mnu_clone.set_sensitive(status)
        self.ui.mnu_delete.set_sensitive(status)
        self.ui.mnu_find.set_sensitive(status)
        self.ui.url_copy.set_sensitive(status)
        self.ui.username_copy.set_sensitive(status)
        self.ui.password_copy.set_sensitive(status)
        self.ui.mnu_preferences.set_sensitive(status)
        self.ui.mnu_open_url.set_sensitive(status)
        self.ui.mnu_info.set_sensitive(status)

        if status is False:
            self.ui.mnu_display_secrets.set_sensitive(False)
        else:
            self.set_show_password_status()

        if status is True and self.needs_saving is True:
            self.ui.mnu_save.set_sensitive(True)
            self.ui.save.set_sensitive(True)
        else:
            self.ui.mnu_save.set_sensitive(False)
            # Work around issue where button is insensitive, but icon is
            # not greyed out
            self.ui.save.set_sensitive(True)
            self.ui.save.set_sensitive(False)

    def set_menu_for_entry(self, status):
        # main menu
        self.ui.mnu_clone.set_sensitive(status)
        self.ui.url_copy.set_sensitive(status)
        self.ui.username_copy.set_sensitive(status)
        self.ui.password_copy.set_sensitive(status)
        self.ui.mnu_open_url.set_sensitive(status)

        # context menu
        self.ui.mnu_clone1.set_sensitive(status)
        self.ui.url_copy1.set_sensitive(status)
        self.ui.username_copy1.set_sensitive(status)
        self.ui.password_copy1.set_sensitive(status)

        # Toolbar
        self.ui.open_url.set_sensitive(status)
        self.ui.copy_username.set_sensitive(status)
        self.ui.copy_password.set_sensitive(status)

        if status is False:
            self.ui.mnu_display_secrets.set_sensitive(False)
            self.ui.display_secrets.set_sensitive(False)
        else:
            self.set_show_password_status()

    def on_add_clicked(self, _toolbutton):
        self.add_entry()

    def on_add_folder_clicked(self, _toolbutton):
        self.add_folder()

    def on_edit_clicked(self, _toolbutton):
        treemodel, treeiter = self.ui.treeview1.get_selection().get_selected()
        if treeiter is not None:
            entry_uuid = treemodel.get_value(treeiter, 2)
            if "pasaffe_treenode." in entry_uuid:
                self.edit_folder(treemodel, treeiter)
            else:
                self.edit_entry(entry_uuid)

    def on_remove_clicked(self, _toolbutton):
        treemodel, treeiter = self.ui.treeview1.get_selection().get_selected()
        if treeiter is not None:
            entry_uuid = treemodel.get_value(treeiter, 2)
            if "pasaffe_treenode." in entry_uuid:
                self.remove_folder()
            else:
                self.remove_entry()

    def set_idle_timeout(self):
        if self.idle_id is not None:
            GLib.source_remove(self.idle_id)
            self.idle_id = None
        if (self.settings.get_boolean('lock-on-idle') is True and
                self.settings.get_int('idle-timeout') != 0):
            idle_time = int(self.settings.get_int('idle-timeout') * 1000 * 60)
            self.idle_id = GLib.timeout_add(
                idle_time, self.idle_timeout_reached)

    def idle_timeout_reached(self):
        if self.is_locked is False:
            self.lock_screen()
        if self.idle_id is not None:
            GLib.source_remove(self.idle_id)
            self.idle_id = None

    def disable_idle_timeout(self):
        if self.idle_id is not None:
            GLib.source_remove(self.idle_id)
            self.idle_id = None

    def set_clipboard_timeout(self):
        if self.clipboard_id is not None:
            GLib.source_remove(self.clipboard_id)
            self.clipboard_id = None
        if self.settings.get_int('clipboard-timeout') != 0:
            clipboard_time = int(
                self.settings.get_int('clipboard-timeout') * 1000)
            self.clipboard_id = GLib.timeout_add(
                clipboard_time, self.clipboard_timeout_reached)

    def clipboard_timeout_reached(self):
        self.clear_clipboard()

    def clear_clipboard(self):
        if self.clipboard_id is not None:
            GLib.source_remove(self.clipboard_id)
            self.clipboard_id = None
        found_copy_in_clipboard = False
        for atom in [Gdk.SELECTION_CLIPBOARD, Gdk.SELECTION_PRIMARY]:
            clipboard = Gtk.Clipboard.get(atom)
            text = clipboard.wait_for_text()
            if text is not None and self.last_copied is not None and \
               text == self.last_copied:
                found_copy_in_clipboard = True
        if found_copy_in_clipboard:
            for atom in [Gdk.SELECTION_CLIPBOARD, Gdk.SELECTION_PRIMARY]:
                clipboard = Gtk.Clipboard.get(atom)
                clipboard.set_text("", 0)
                clipboard.store()
        self.last_copied = None

    def set_show_password_status(self):
        visible = self.settings.get_boolean('visible-secrets')

        if visible is True:
            self.ui.mnu_display_secrets.set_sensitive(False)
            # Work around issue where button is insensitive, but icon is
            # not greyed out
            self.ui.display_secrets.set_sensitive(True)
            self.ui.display_secrets.set_sensitive(False)
        else:
            self.ui.mnu_display_secrets.set_sensitive(True)
            self.ui.display_secrets.set_sensitive(True)

    def _set_title(self):
        prefix = ""
        if not self.default_database:
            prefix = "%s - " % os.path.basename(self.database)
        self.set_title("%s%sPasaffe" % (
            prefix,
            "*" if self.needs_saving else ""))

    def set_save_status(self, needed):
        self.needs_saving = needed
        self.ui.save.set_sensitive(needed)
        self.ui.mnu_save.set_sensitive(needed)
        self._set_title()

    def get_save_status(self):
        return self.needs_saving



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2011-2013 Marc Deslauriers <marc.deslauriers@canonical.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gio, Gtk  # noqa: E402
from gi.repository import Gdk  # noqa: E402

from pasaffe_lib.helpersgui import get_builder  # noqa: E402
from pasaffe_lib.helpers import gen_password  # noqa: E402

# pylint: disable=E1101


class EditDetailsDialog(Gtk.Dialog):
    __gtype_name__ = "EditDetailsDialog"

    def __new__(cls):
        """Special static method that's automatically called by Python when
        constructing a new instance of this class.

        Returns a fully instantiated EditDetailsDialog object.
        """
        builder = get_builder('EditDetailsDialog')
        new_object = builder.get_object('edit_details_dialog')
        new_object.finish_initializing(builder)
        return new_object

    def finish_initializing(self, builder):
        """Called when we're finished initializing.

        finish_initalizing should be called after parsing the ui definition
        and creating a EditDetailsDialog object with it in order to
        finish initializing the start of the new EditDetailsDialog
        instance.
        """
        # Get a reference to the builder and set up the signals.
        self.builder = builder
        self.ui = builder.get_ui(self)

        settings = Gio.Settings.new("net.launchpad.pasaffe")
        self.password_length = settings.get_int("password-length")

    def on_btn_ok_clicked(self, widget, data=None):
        """The user has elected to save the changes.

        Called before the dialog returns Gtk.RESONSE_OK from run().
        """
        pass

    def on_btn_cancel_clicked(self, widget, data=None):
        """The user has elected cancel changes.

        Called before the dialog returns Gtk.ResponseType.CANCEL for run()
        """
        pass

    def on_password_button_clicked(self, widget):
        """The user has clicked the password button"""
        self.show_passwords_menu(widget)

    def on_menuitem_activate(self, widget):
        """The user has clicked on a menu item"""
        label = widget.get_label()
        self.ui.password_entry.set_text(label)

    def show_passwords_menu(self, widget):
        """Generate some new passwords"""
        try:
            passwords = gen_password(6, self.password_length)
            self.ui.password1.set_label(passwords[0].decode('utf-8'))
            self.ui.password2.set_label(passwords[1].decode('utf-8'))
            self.ui.password3.set_label(passwords[2].decode('utf-8'))
            self.ui.password4.set_label(passwords[3].decode('utf-8'))
            self.ui.password5.set_label(passwords[4].decode('utf-8'))
            self.ui.password6.set_label(passwords[5].decode('utf-8'))

        except:  # noqa: E722
            pass

        if not Gtk.check_version(3, 22, 0):
            self.ui.password_menu.popup_at_widget(widget,
                                                  Gdk.Gravity.SOUTH_WEST,
                                                  Gdk.Gravity.NORTH_WEST,
                                                  None)
        else:
            self.ui.password_menu.popup(None, None, None, None,
                                        0, Gtk.get_current_event_time())


if __name__ == "__main__":
    dialog = EditDetailsDialog()
    dialog.show()
    Gtk.main()



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2013 Marc Deslauriers <marc.deslauriers@canonical.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import os.path
import unittest
sys.path.insert(0, os.path.realpath(os.path.join(os.path.dirname(__file__),
                                                 "..")))

from pasaffe_lib.readdb import PassSafeFile  # noqa: E402


class TestPasswordSafe331(unittest.TestCase):
    def setUp(self):
        self.passfile = PassSafeFile('./tests/databases/pwsafe-331.psafe3',
                                     'pasaffe')

    def test_num_entries(self):
        self.assertEqual(len(self.passfile.records), 3)

    def test_empty_folders(self):

        empty_folders = [['emptygroup1'],
                         ['emptygroup1', 'emptygroup2'],
                         ['emptygroup1', 'emptygroup2', 'emptygroup3'],
                         ['level1group', 'level2group'],
                         ['emptygroup1', 'test'],
                         ['emptygroup1', 'with/slash']]

        empty_fields = ['emptygroup1',
                        'emptygroup1.emptygroup2',
                        'emptygroup1.emptygroup2.emptygroup3',
                        'level1group.level2group',
                        'emptygroup1.test',
                        'emptygroup1.with/slash']

        self.assertEqual(len(self.passfile.empty_folders), len(empty_fields))
        self.assertEqual(self.passfile.get_empty_folders(), empty_folders)
        self.assertEqual(self.passfile.empty_folders, empty_fields)

    def test_get_database_version_string(self):
        self.assertEqual(self.passfile.get_database_version_string(), "03.0b")

    def test_get_database_uuid(self):
        self.assertEqual(self.passfile.header[1],
                         b'cf\xfe\xea\x00wBU\xa2TM\xc3k\x0f>\x0f')

    def test_get_saved_name(self):
        self.assertEqual(self.passfile.get_saved_name(), "mdeslaur")

    def test_get_saved_host(self):
        self.assertEqual(self.passfile.get_saved_host(), "mdlinux")

    def test_get_saved_application(self):
        self.assertEqual(self.passfile.get_saved_application(),
                         'Password Safe V3.31')

    def test_get_saved_date_string(self):
        self.assertEqual(self.passfile.get_saved_date_string(False),
                         'Thu, 25 Jul 2013 23:57:08')

    def test_entry_1(self):
        uuid = '4a32a8ad616343b692e85c721bfce0e2'
        self.assertFalse(2 in self.passfile.records[uuid])
        self.assertEqual(self.passfile.get_folder_list(uuid), [])
        self.assertEqual(self.passfile.records[uuid][3], 'topentry1')
        self.assertEqual(self.passfile.records[uuid][4], 'username1')
        self.assertEqual(self.passfile.records[uuid][5],
                         'This is a note\nThis is a second line\n'
                         'Unicode: éléphant')
        self.assertEqual(self.passfile.records[uuid][6], 'password1')
        self.assertEqual(self.passfile.get_creation_time(uuid, False),
                         'Thu, 25 Jul 2013 00:21:00')
        self.assertEqual(self.passfile.records[uuid][13],
                         'http://www.example.com')

    def test_entry_2(self):
        uuid = '722328d418584201803a119fa517b799'
        self.assertEqual(self.passfile.records[uuid][2], 'level1group')
        self.assertEqual(self.passfile.get_folder_list(uuid), ['level1group'])
        self.assertEqual(self.passfile.records[uuid][3], 'level1entry')
        self.assertEqual(self.passfile.records[uuid][4], 'username1')
        self.assertEqual(self.passfile.records[uuid][5], 'This is a note\n')
        self.assertEqual(self.passfile.records[uuid][6], 'password1')
        self.assertEqual(self.passfile.get_creation_time(uuid, False),
                         'Thu, 25 Jul 2013 00:25:42')

    def test_entry_3(self):
        uuid = 'cb16d230853247ad8cb12ef6ea615cb4'
        self.assertEqual(self.passfile.records[uuid][2],
                         'level1group.level2group.level3group')
        self.assertEqual(self.passfile.get_folder_list(uuid),
                         ['level1group', 'level2group', 'level3group'])
        self.assertEqual(self.passfile.records[uuid][3], 'level3entry')
        self.assertEqual(self.passfile.records[uuid][4], 'usernamelevel3')
        self.assertFalse(5 in self.passfile.records[uuid])
        self.assertEqual(self.passfile.records[uuid][6], 'passwordlevel3')
        self.assertEqual(self.passfile.get_creation_time(uuid, False),
                         'Thu, 25 Jul 2013 00:26:36')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2013 Marc Deslauriers <marc.deslauriers@canonical.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import os.path
sys.path.insert(0, os.path.realpath(os.path.join(os.path.dirname(__file__),
                                                 "..")))

from pasaffe_lib.readdb import PassSafeFile  # noqa: E402


def get_test_data():
    empty_folders = [['emptygroup1'],
                     ['emptygroup1', 'emptygroup2'],
                     ['emptygroup1', 'emptygroup2', 'emptygroup3'],
                     ['level1group', 'level2group'],
                     ['emptygroup1', 'test'],
                     ['emptygroup1', 'with/slash']]

    entries = [{3: "topentry1",
                4: "username1",
                5: "This is a note",
                6: "password1",
                13: "http://www.example.com"},
               {2: "level1group",
                3: "level1entry",
                4: "username1",
                5: "This is a note",
                6: "password1"},
               {2: "level1group.level2group.level3group",
                3: "level3entry",
                4: "usernamelevel3",
                6: "passwordlevel3"}]

    return entries, empty_folders


def create_test_db():
    passfile = PassSafeFile()
    passfile.new_db('pasaffe')

    entries, empty_folders = get_test_data()

    for entry in entries:
        uuid = passfile.new_entry()
        for key in entry.keys():
            passfile.records[uuid][key] = entry[key]

    for folder in empty_folders:
        passfile.add_empty_folder(folder)

    return passfile


if __name__ == '__main__':
    # This script can be used to create a test database
    # Simply run it with a filename, and a database will be created

    if len(sys.argv) < 2:
        print("Missing database output filename. Aborting.",
              file=sys.stderr)
        sys.exit(1)

    filename = sys.argv[1]
    if os.path.exists(filename):
        print("Output file '%s' already exists. Aborting." % filename,
              file=sys.stderr)
        sys.exit(1)

    passfile = create_test_db()
    passfile.writefile(filename)



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2011-2013 Marc Deslauriers <marc.deslauriers@canonical.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""GUI helpers for an Ubuntu application."""
import os

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk  # noqa: E402

from . Builder import Builder  # noqa: E402
from . pasaffeconfig import get_data_file  # noqa: E402


def get_builder(builder_file_name):
    """Return a fully-instantiated Gtk.Builder instance from specified ui
    file

    :param builder_file_name: The name of the builder file, without extension.
        Assumed to be in the 'ui' directory under the data path.
    """
    # Look for the ui file that describes the user interface.
    ui_filename = get_data_file('ui', '%s.ui' % (builder_file_name,))
    if not os.path.exists(ui_filename):
        ui_filename = None

    builder = Builder()
    builder.set_translation_domain('pasaffe')
    builder.add_from_file(ui_filename)
    return builder


def show_uri(parent, link):
    screen = parent.get_screen()
    try:
        Gtk.show_uri(screen, link, Gtk.get_current_event_time())
    except:  # noqa: E722
        pass



# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2011-2013 Marc Deslauriers <marc.deslauriers@canonical.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk  # noqa: E402

from pasaffe_lib.helpersgui import get_builder  # noqa: E402


class NewDatabaseDialog(Gtk.Dialog):
    __gtype_name__ = "NewDatabaseDialog"

    def __new__(cls):
        """Special static method that's automatically called by Python when
        constructing a new instance of this class.

        Returns a fully instantiated NewDatabaseDialog object.
        """
        builder = get_builder('NewDatabaseDialog')
        new_object = builder.get_object('new_database_dialog')
        new_object.finish_initializing(builder)
        return new_object

    def finish_initializing(self, builder):
        """Called when we're finished initializing.

        finish_initalizing should be called after parsing the ui definition
        and creating a NewDatabaseDialog object with it in order to
        finish initializing the start of the new NewDatabaseDialog
        instance.
        """
        # Get a reference to the builder and set up the signals.
        self.builder = builder
        self.ui = builder.get_ui(self)

    def on_btn_ok_clicked(self, widget, data=None):
        """The user has elected to save the changes.

        Called before the dialog returns Gtk.RESONSE_OK from run().
        """
        pass

    def on_btn_cancel_clicked(self, widget, data=None):
        """The user has elected cancel changes.

        Called before the dialog returns Gtk.ResponseType.CANCEL for run()
        """
        pass


if __name__ == "__main__":
    dialog = NewDatabaseDialog()
    dialog.show()
    Gtk.main()
